"""
pygame-menu
https://github.com/ppizarror/pygame-menu

FRAME
Widget container.
"""

__all__ = [

    # Main class
    'Frame',
    'FrameManager',

    # Types
    'FrameTitleBackgroundColorType',
    'FrameTitleButtonType',

    # Constants
    'FRAME_DEFAULT_TITLE_BACKGROUND_COLOR',
    'FRAME_TITLE_BUTTON_CLOSE',
    'FRAME_TITLE_BUTTON_MAXIMIZE',
    'FRAME_TITLE_BUTTON_MINIMIZE'

]

import pygame
import pygame_menu

from abc import ABC
from pygame_menu._decorator import Decorator
from pygame_menu.baseimage import BaseImage
from pygame_menu.locals import CURSOR_HAND, ORIENTATION_VERTICAL, \
    ORIENTATION_HORIZONTAL, ALIGN_CENTER, ALIGN_LEFT, ALIGN_RIGHT, POSITION_CENTER, \
    POSITION_NORTH, POSITION_SOUTH, FINGERUP, FINGERDOWN, FINGERMOTION
from pygame_menu.font import FontType, assert_font
from pygame_menu.utils import assert_alignment, make_surface, assert_vector, \
    assert_orientation, assert_color, fill_gradient, parse_padding, uuid4, warn, \
    get_finger_pos
from pygame_menu.widgets.core.widget import Widget, check_widget_mouseleave, \
    WidgetTransformationNotImplemented, AbstractWidgetManager
from pygame_menu.widgets.widget.button import Button
from pygame_menu.widgets.widget.label import Label

from pygame_menu._types import Optional, NumberType, Dict, Tuple, Union, List, \
    Vector2NumberType, Tuple2IntType, NumberInstance, Any, ColorInputType, \
    EventVectorType, PaddingType, CallbackType, ColorInputGradientType, \
    CursorInputType, VectorInstance

# Constants
FRAME_DEFAULT_TITLE_BACKGROUND_COLOR = ((10, 36, 106), (166, 202, 240), False, True)
FRAME_TITLE_BUTTON_CLOSE = 'close'
FRAME_TITLE_BUTTON_MAXIMIZE = 'maximize'
FRAME_TITLE_BUTTON_MINIMIZE = 'minimize'
S_FINGER_FACTOR = 0.25, 0.25

# Types
FrameTitleBackgroundColorType = Optional[Union[ColorInputType, ColorInputGradientType, BaseImage]]
FrameTitleButtonType = str


# noinspection PyMissingOrEmptyDocstring,PyProtectedMember
class Frame(Widget):
    """
    Frame is a widget container, it can pack many widgets.

    All widgets inside have a floating position. Widgets inside are placed using
    its margin + width/height. ``(0, 0)`` coordinate is the top-left position in
    frame.

    Frame modifies ``translation`` of widgets. Thus, if widget has a translation
    before it will be removed.

    Widget packing currently only supports LEFT, CENTER and RIGHT alignment in
    the same line, widgets cannot be packed in two different lines. For such
    purpose use several frames. If widget width or height exceeds Frame size
    ``_FrameSizeException`` will be raised.

    .. note::

        Frame cannot be selected. Thus, it does not receive any selection effect.

    .. note::

        Frames should be appended to Menu scrollable frames if it's scrollable,
        be careful when removing. Check :py:class:`pygame_menu.widgets.DropSelect`
        as a complete example of Frame implementation within another Widgets.

    .. note::

        Frame only accepts translation and resize transformations.

    :param width: Frame width in px
    :param height: Frame height in px
    :param orientation: Frame orientation (horizontal or vertical). See :py:mod:`pygame_menu.locals`
    :param frame_id: ID of the frame
    """
    _accepts_scrollarea: bool
    _accepts_title: bool
    _control_widget: Optional['Widget']
    _control_widget_last_pos: Optional[Vector2NumberType]
    _draggable: bool
    _frame_scrollarea: Optional['pygame_menu._scrollarea.ScrollArea']
    _frame_size: Tuple2IntType
    _frame_title: Optional['Frame']
    _has_frames: bool  # True if frame has packed other frames
    _has_title: bool
    _height: int
    _menu_can_be_none_pack: bool
    _orientation: str
    _pack_margin_warning: bool
    _pos: Dict[str, Tuple[int, int]]  # Widget positioning
    _real_rect: 'pygame.Rect'
    _recursive_render: int
    _widgets: Dict[str, 'Widget']  # widget
    _widgets_props: Dict[str, Tuple[str, str]]  # alignment, vertical position
    _width: int
    first_index: int  # First selectable widget index
    horizontal: bool
    last_index: int  # Last selectable widget index

    def __init__(
        self,
        width: NumberType,
        height: NumberType,
        orientation: str,
        frame_id: str = ''
    ) -> None:
        super(Frame, self).__init__(widget_id=frame_id)
        assert isinstance(width, NumberInstance)
        assert isinstance(height, NumberInstance)
        assert width > 0, \
            f'width must be greater than zero ({width} received)'
        assert height > 0, \
            f'height must be greater than zero ({height} received)'
        assert_orientation(orientation)

        # Internals
        self._accepts_scrollarea = True
        self._accepts_title = True
        self._control_widget = None
        self._control_widget_last_pos = None  # This checks if menu has updated widget position
        self._draggable = False
        self._frame_scrollarea = None
        self._frame_size = (width, height)  # Size of the frame, set in make_scrollarea
        self._has_frames = False
        self._height = int(height)
        self._menu_can_be_none_pack = False
        self._orientation = orientation
        self._pack_margin_warning = True  # Set to False for hiding the pack margin warning
        self._pos = {}
        self._real_rect = pygame.Rect(0, 0, width, height)
        self._recursive_render = 0
        self._relax = False  # If True ignore sizing
        self._widgets = {}
        self._widgets_props = {}
        self._width = int(width)

        # Title
        self._frame_title = None
        self._has_title = False

        # Configure widget public's
        self.first_index = -1
        self.horizontal = orientation == ORIENTATION_HORIZONTAL
        self.is_scrollable = False
        self.is_selectable = False
        self.last_index = -1

    def set_title(
        self,
        title: str,
        cursor: CursorInputType = None,
        background_color: FrameTitleBackgroundColorType = FRAME_DEFAULT_TITLE_BACKGROUND_COLOR,
        draggable: bool = False,
        padding_inner: PaddingType = 0,
        padding_outer: PaddingType = 0,
        title_alignment: str = ALIGN_LEFT,
        title_buttons_alignment: str = ALIGN_RIGHT,
        title_font: Optional[FontType] = None,
        title_font_color: Optional[ColorInputType] = None,
        title_font_size: Optional[int] = None
    ) -> 'Frame':
        """
        Add a title to the frame.

        :param title: New title
        :param cursor: Title cursor
        :param background_color: Title background color. It can be a Color, a gradient, or an image
        :param draggable: If ``True`` the title accepts user drag using the mouse
        :param padding_inner: Padding inside the title
        :param padding_outer: Padding outside the title (respect to the Frame)
        :param title_alignment: Alignment of the title
        :param title_buttons_alignment: Alignment of the title buttons (if appended later)
        :param title_font: Title font. If ``None`` uses the same as the Frame
        :param title_font_color: Title font color. If ``None`` uses the same as the Frame
        :param title_font_size: Title font size in px. If ``None`` uses the same as the Frame
        :return: Title frame object
        """
        assert self.configured, \
            f'{self.get_class_id()} must be configured before setting a title'
        if not self._accepts_title:
            raise _FrameDoNotAcceptTitle(f'{self.get_class_id()} does not accept a title')
        self._title = title

        # If it has previous title
        self.remove_title()

        assert isinstance(draggable, bool)
        self._draggable = draggable

        # Format title font properties
        if title_font is None:
            title_font = self._font_name
        assert_font(title_font)
        if title_font_color is None:
            title_font_color = self._font_color
        title_font_color = assert_color(title_font_color)
        if title_font_size is None:
            title_font_size = self._font_size
        assert isinstance(title_font_size, int) and title_font_size > 0
        assert_alignment(title_alignment)
        assert_alignment(title_buttons_alignment)

        # Check alignment are different
        assert title_alignment != title_buttons_alignment, \
            'title alignment and buttons alignment must be different'

        # Create title widget
        title_label = Label(
            title=title,
            label_id=self._id + '+title+label-' + uuid4(short=True)
        )
        title_label.set_font(
            antialias=self._font_antialias,
            background_color=None,
            color=title_font_color,
            font=title_font,
            font_size=title_font_size,
            readonly_color=self._font_readonly_color,
            readonly_selected_color=self._font_readonly_selected_color,
            selected_color=self._font_selected_color
        )
        title_label.set_tab_size(self._tab_size)
        title_label.configured = True
        title_label.set_menu(self._menu)
        title_label._update__repr___(self)

        # Create frame title
        pad_outer = parse_padding(padding_outer)  # top, right, bottom, left
        pad_inner = parse_padding(padding_inner)
        self._frame_title = Frame(
            width=self.get_width() - (pad_outer[1] + pad_outer[3] + pad_inner[1] + pad_inner[3]),
            height=title_label.get_height(),
            orientation=ORIENTATION_HORIZONTAL,
            frame_id=self._id + '+title-' + uuid4(short=True)
        )
        self._frame_title._accepts_scrollarea = False
        self._frame_title._accepts_title = False
        self._frame_title._menu_can_be_none_pack = True
        self._frame_title._menu = self._menu
        self._frame_title.set_attribute('buttons_alignment', title_buttons_alignment)
        self._frame_title.set_attribute('pbottom', pad_outer[2] - pad_inner[2])
        self._frame_title.set_cursor(cursor if cursor is not None else self._cursor)
        self._frame_title.set_padding(padding_inner)
        self._frame_title.set_scrollarea(self._scrollarea)
        self._frame_title.translate(pad_outer[3] + pad_inner[3], pad_outer[0] + pad_inner[0])
        self._frame_title.set_controls(
            joystick=self._joystick_enabled,
            mouse=self._mouse_enabled,
            touchscreen=self._touchscreen_enabled,
            keyboard=self._keyboard_enabled
        )
        if self._frame is not None:
            self._frame_title.set_frame(self._frame)
        self._frame_title._update__repr___(self)

        # Store constructor
        self._frame_title.set_attribute(
            'constructor', {
                'title': title,
                'cursor': cursor,
                'background_color': background_color,
                'draggable': draggable,
                'padding_inner': padding_inner,
                'padding_outer': padding_outer,
                'title_alignment': title_alignment,
                'title_buttons_alignment': title_buttons_alignment,
                'title_font': title_font,
                'title_font_color': title_font_color,
                'title_font_size': title_font_size
            }
        )

        # Create frame title background rect
        title_bg = make_surface(
            self.get_width(),
            title_label.get_height() + pad_outer[0] + pad_outer[2] + pad_inner[0] + pad_inner[2]
        )

        # Blit frame bgrect if scrollable
        if self._frame_scrollarea is not None and self.is_scrollable:
            area_color = self._background_color
            if isinstance(area_color, BaseImage):
                area_color.draw(title_bg, area=title_bg.get_rect())
            elif area_color is not None:
                title_bg.fill(area_color, rect=title_bg.get_rect())
        self._frame_title.get_decorator().add_surface(-title_bg.get_width() / 2, -title_bg.get_height() / 2 + 1, title_bg)

        # Set background
        is_color = True
        try:
            assert_color(background_color, warn_if_invalid=False)
        except (ValueError, AssertionError):
            is_color = False
        if isinstance(background_color, pygame_menu.BaseImage) or is_color:
            self._frame_title.set_background_color(background_color)
        else:  # Is gradient
            assert isinstance(background_color, tuple)
            assert len(background_color) == 4, \
                'gradient color type must has 4 components (from color, to color, ' \
                'vertical, forward)'
            w, h = self._frame_title.get_size()
            new_surface = make_surface(w, h)
            fill_gradient(
                surface=new_surface,
                color=background_color[0],
                gradient=background_color[1],
                vertical=background_color[2],
                forward=background_color[3]
            )
            self._frame_title.get_decorator().add_surface(-w / 2, -h / 2 + 1, new_surface)

        # Pack title
        self._frame_title.pack(title_label, align=title_alignment)

        self._has_title = True
        self._render()
        self.set_position(self._position[0], self._position[1])
        self.force_menu_surface_update()

        # Title adds frame to scrollable frames even if not scrollable
        self._append_menu_update_frame(self)

        return self._frame_title

    def remove_title(self) -> 'Frame':
        """
        Remove title from current Frame.

        :return: Self reference
        """
        if not self._accepts_title:
            raise _FrameDoNotAcceptTitle(f'{self.get_class_id()} does not accept a title')
        if self._has_title:
            self._frame_title = None
            self._has_title = False
        if not self.is_scrollable:
            self._remove_menu_update_frame(self)
        self._draggable = False
        self._render()
        self.force_menu_surface_update()
        return self

    def add_title_generic_button(
        self,
        button: 'Button',
        margin: Vector2NumberType = (0, 0)
    ) -> 'Frame':
        """
        Add button to title. Button kwargs receive the ``button`` reference and
        the Frame reference in ``frame`` argument, such as:

        .. code-block:: python

            onreturn_button_callback(*args, frame=Frame, button=Button, **kwargs)

        :param button: Button object
        :param margin: Pack margin on x-axis and y-axis (x, y) in px
        :return: Self reference
        """
        if not self._accepts_title:
            raise _FrameDoNotAcceptTitle(f'{self.get_class_id()} does not accept a title')
        assert self._has_title, \
            f'{self.get_class_id()} does not have any title, call set_title(...) beforehand'
        assert isinstance(button, Button)
        assert button.get_menu() is None, \
            f'{button.get_class_id()} menu reference must be None'
        assert button.configured, \
            f'{button.get_class_id()} must be configured before addition to title'
        assert button.get_frame() is None, \
            f'{button.get_class_id()} frame must be None'

        # Check sizing
        total_title_height = self._frame_title.get_height(apply_padding=False)
        assert button.get_height() <= total_title_height, \
            f'{button.get_class_id()} height ({button.get_height()}) must be lower' \
            f' than frame title height ({total_title_height})'

        # Add frame to button kwargs
        if 'frame' in button._kwargs.keys():
            raise ValueError(f'{button.get_class_id()} already has "frame" kwargs option')
        button._kwargs['frame'] = self
        if 'button' in button._kwargs.keys():
            raise ValueError(f'{button.get_class_id()} already has "button" kwargs option')
        button._kwargs['button'] = button

        # Pack
        align = self._frame_title.get_attribute('buttons_alignment')
        button.set_attribute('align', align)
        button.set_attribute('margin', margin)
        self._frame_title.pack(button, align=align, margin=margin)
        self._frame_title.update_position()

        return self

    def add_title_button(
        self,
        style: FrameTitleButtonType,
        callback: CallbackType,
        background_color: ColorInputType = (150, 150, 150),
        cursor: CursorInputType = CURSOR_HAND,
        margin: Vector2NumberType = (4, 0),
        symbol_color: ColorInputType = (0, 0, 0),
        symbol_height: NumberType = 0.75,
        symbol_margin: int = 4
    ) -> 'Button':
        """
        Add predefined button to title. The button kwargs receive the ``button``
        reference and the Frame reference in ``frame`` argument, such as:

        .. code-block:: python

            onreturn_button_callback(*args, frame=Frame, button=Button, **kwargs)

        :param style: Style of the button (changes the symbol)
        :param callback: Callback of the button if pressed
        :param cursor: Button cursor
        :param background_color: Button background color
        :param margin: Pack margin on x-axis and y-axis (x, y) in px
        :param symbol_color: Color of the symbol
        :param symbol_height: Symbol height factor, if ``1.0`` uses 100% of the button height
        :param symbol_margin: Symbol margin in px
        :return: Added button
        """
        if not self._accepts_title:
            raise _FrameDoNotAcceptTitle(f'{self.get_class_id()} does not accept a title')
        assert self._has_title, \
            f'{self.get_class_id()} does not have any title, call set_title(...) beforehand'
        assert isinstance(symbol_height, NumberInstance) and 0 <= symbol_height <= 1
        assert isinstance(symbol_margin, int) and 0 <= symbol_margin
        h = self._frame_title.get_height(apply_padding=False) * symbol_height
        dh = self._frame_title.get_height(apply_padding=False) * (1 - symbol_height)
        assert symbol_margin < h / 2
        if dh > 0:
            dh += 1

        # Create button
        btn = Button('', onreturn=callback,
                     button_id=self._frame_title._id + '+button-' + uuid4(short=True))
        btn.set_padding(h / 2)
        btn.translate(0, dh / 2)
        btn.set_cursor(cursor)
        btn.set_background_color(background_color)
        btn.configured = True
        btn._update__repr___(self)

        # Create style decoration
        btn_rect = btn.get_rect()
        btn_rect.x = 0
        btn_rect.y = 0
        t = symbol_margin
        border = 1

        if style == FRAME_TITLE_BUTTON_CLOSE:
            style_pos = (
                (btn_rect.left + t, btn_rect.top + t),
                (btn_rect.centerx, btn_rect.centery),
                (btn_rect.right - t, btn_rect.top + t),
                (btn_rect.centerx, btn_rect.centery),
                (btn_rect.right - t, btn_rect.bottom - t),
                (btn_rect.centerx, btn_rect.centery),
                (btn_rect.left + t, btn_rect.bottom - t),
                (btn_rect.centerx, btn_rect.centery),
                (btn_rect.left + t, btn_rect.top + t)
            )
            border = 0

        elif style == FRAME_TITLE_BUTTON_MAXIMIZE:
            style_pos = (
                (btn_rect.left + t, btn_rect.bottom - t),
                (btn_rect.right - t, btn_rect.bottom - t),
                (btn_rect.right - t, btn_rect.top + t),
                (btn_rect.left + t, btn_rect.top + t)
            )

        elif style == FRAME_TITLE_BUTTON_MINIMIZE:
            style_pos = (
                (btn_rect.left + t, btn_rect.centery + border),
                (btn_rect.right - t, btn_rect.centery + border)
            )

        else:
            raise ValueError(f'unknown button style "{style}"')

        # Draw style
        style_surface = make_surface(h, h, alpha=True)
        # noinspection PyArgumentList
        pygame.draw.polygon(style_surface, symbol_color, style_pos, border)
        btn.get_decorator().add_surface(0, 0, surface=style_surface, centered=True)

        self.add_title_generic_button(btn, margin)
        return btn

    def get_title(self) -> str:
        if not self._has_title:
            # raise ValueError(f'{self.get_class_id()} does not have any title')
            return ''
        return self._title

    def get_inner_size(self) -> Tuple2IntType:
        """
        Return Frame inner size (width, height).

        :return: Size tuple in px
        """
        return self._width, self._height

    def _get_menu_update_frames(self) -> List['pygame_menu.widgets.Frame']:
        """
        Return the menu update frames list.

        .. warning::

            Use with caution.

        :return: Frame update list if the menu reference is not ``None``, else, return an empty list
        """
        if self._menu is not None:
            return self._menu._update_frames
        return []

    def _sort_menu_update_frames(self) -> None:
        """
        Sort the menu update frames (frames which receive updates).
        """
        if self._menu is not None:
            self._menu._sort_update_frames()

    def _append_menu_update_frame(self, frame: 'Frame') -> None:
        """
        Append update frame to menu and sort.

        :param frame: Frame to append
        """
        assert isinstance(frame, Frame)
        update_frames = self._get_menu_update_frames()
        if frame not in update_frames:
            update_frames.append(frame)
            self._sort_menu_update_frames()

    def _remove_menu_update_frame(self, frame: 'Frame') -> None:
        """
        Remove update frame to menu and sort.

        :param frame: Frame to append
        """
        assert isinstance(frame, Frame)
        update_frames = self._get_menu_update_frames()
        if frame in update_frames:
            update_frames.remove(frame)

    def on_remove_from_menu(self) -> 'Frame':
        for w in self.get_widgets(unpack_subframes=False):
            self.unpack(w)
        self.update_indices()
        return self

    def set_menu(self, menu: Optional['pygame_menu.Menu']) -> 'Frame':
        # If menu is set, remove from previous scrollable if enabled
        self._remove_menu_update_frame(self)

        # Update menu
        super(Frame, self).set_menu(menu)

        # Add self to scrollable
        if self.is_scrollable:
            self._append_menu_update_frame(self)

        return self

    def relax(self, relax: bool = True) -> 'Frame':
        """
        Set relax status. If ``True`` Frame ignores sizing checks.

        :param relax: Relax status
        :return: Self reference
        """
        assert isinstance(relax, bool)
        self._relax = relax
        return self

    def get_max_size(self) -> Tuple2IntType:
        """
        Return the max size of the frame.

        :return: Max (width, height) in px
        """
        if self._frame_scrollarea is not None:
            return self._frame_scrollarea.get_size(inner=True)
        return self.get_size()

    def make_scrollarea(
        self,
        max_width: Optional[NumberType],
        max_height: Optional[NumberType],
        scrollarea_color: Optional[Union[ColorInputType, 'pygame_menu.BaseImage']],
        scrollbar_color: ColorInputType,
        scrollbar_cursor: CursorInputType,
        scrollbar_shadow: bool,
        scrollbar_shadow_color: ColorInputType,
        scrollbar_shadow_offset: int,
        scrollbar_shadow_position: str,
        scrollbar_slider_color: ColorInputType,
        scrollbar_slider_hover_color: ColorInputType,
        scrollbar_slider_pad: NumberType,
        scrollbar_thick: NumberType,
        scrollbars: Union[str, Tuple[str, ...]]
    ) -> 'Frame':
        """
        Make the scrollarea of the frame.

        :param max_width: Maximum width of the scrollarea in px
        :param max_height: Maximum height of the scrollarea in px
        :param scrollarea_color: Scroll area color or image. If ``None`` area is transparent
        :param scrollbar_color: Scrollbar color
        :param scrollbar_cursor: Scrollbar cursor
        :param scrollbar_shadow: Indicate if a shadow is drawn on each scrollbar
        :param scrollbar_shadow_color: Color of the shadow of each scrollbar
        :param scrollbar_shadow_offset: Offset of the scrollbar shadow in px
        :param scrollbar_shadow_position: Position of the scrollbar shadow. See :py:mod:`pygame_menu.locals`
        :param scrollbar_slider_color: Color of the sliders
        :param scrollbar_slider_hover_color: Color of the slider if hovered or clicked
        :param scrollbar_slider_pad: Space between slider and scrollbars borders in px
        :param scrollbar_thick: Scrollbar thickness in px
        :param scrollbars: Positions of the scrollbars. See :py:mod:`pygame_menu.locals`
        :return: Self reference
        """
        if not self._accepts_scrollarea:
            raise _FrameDoNotAcceptScrollarea(f'{self.get_class_id()} does not accept a scrollarea')
        assert len(self._widgets.keys()) == 0, 'frame widgets must be empty if creating the scrollarea'
        assert self.configured, 'frame must be configured before adding the scrollarea'
        if max_width is None:
            max_width = self._width
        if max_height is None:
            max_height = self._height
        assert isinstance(max_width, NumberInstance)
        assert isinstance(max_height, NumberInstance)
        assert 0 < max_width <= self._width, \
            f'scroll area width ({max_width}) cannot exceed frame width ({self._width})'
        assert 0 < max_height <= self._height, \
            f'scroll area height ({max_height}) cannot exceed frame height ({self._height})'
        # if not self._relax:
        #     pass
        # else:
        #     max_height = min(max_height, self._height)
        #     max_width = min(max_width, self._width)

        sx = 0 if self._width == max_width else scrollbar_thick
        sy = 0 if self._height == max_height else scrollbar_thick

        if self._width > max_width or self._height > max_height:
            self.is_scrollable = True
            self.set_padding(0)
        else:
            # Configure size
            self._frame_size = (self._width, self._height)
            self._frame_scrollarea = None

            # If in previous scrollable frames
            if self.is_scrollable:
                self._remove_menu_update_frame(self)

            self.is_scrollable = False
            return self

        # Create area object
        self._frame_scrollarea = pygame_menu._scrollarea.ScrollArea(
            area_color=scrollarea_color,
            area_height=max_height + sx,
            area_width=max_width + sy,
            controls_joystick=self._joystick_enabled,
            controls_keyboard=self._keyboard_enabled,
            controls_mouse=self._mouse_enabled,
            controls_touchscreen=self._touchscreen_enabled,
            parent_scrollarea=self._scrollarea,
            scrollbar_color=scrollbar_color,
            scrollbar_cursor=scrollbar_cursor,
            scrollbar_slider_color=scrollbar_slider_color,
            scrollbar_slider_hover_color=scrollbar_slider_hover_color,
            scrollbar_slider_pad=scrollbar_slider_pad,
            scrollbar_thick=scrollbar_thick,
            scrollbars=scrollbars,
            shadow=scrollbar_shadow,
            shadow_color=scrollbar_shadow_color,
            shadow_offset=scrollbar_shadow_offset,
            shadow_position=scrollbar_shadow_position
        )

        # Store constructor data
        self._frame_scrollarea.set_attribute(
            'constructor',
            {
                'scrollarea_color': scrollarea_color,
                'scrollbar_color': scrollbar_color,
                'scrollbar_cursor': scrollbar_cursor,
                'scrollbar_shadow': scrollbar_shadow,
                'scrollbar_shadow_color': scrollbar_shadow_color,
                'scrollbar_shadow_offset': scrollbar_shadow_offset,
                'scrollbar_shadow_position': scrollbar_shadow_position,
                'scrollbar_slider_color': scrollbar_slider_color,
                'scrollbar_slider_hover_color': scrollbar_slider_hover_color,
                'scrollbar_slider_pad': scrollbar_slider_pad,
                'scrollbar_thick': scrollbar_thick,
                'scrollbars': scrollbars
            }
        )

        if self._width == max_width:
            self._frame_scrollarea.hide_scrollbars(ORIENTATION_HORIZONTAL)
        if self._height == max_height:
            self._frame_scrollarea.hide_scrollbars(ORIENTATION_VERTICAL)

        # Create surface
        self._surface = make_surface(self._width, self._height, alpha=True)

        # Configure area
        self._frame_scrollarea.set_world(self._surface)

        # Configure size
        self._frame_size = (max_width + sy, max_height + sx)

        # Set menu
        self._frame_scrollarea.set_menu(self._menu)

        # If it has title
        if self._has_title:
            if self._verbose:
                warn(f'previous {self.get_class_id()} title has been removed')
            self.remove_title()

        return self

    def get_indices(self) -> Tuple[int, int]:
        """
        Return first and last selectable indices tuple.

        :return: First, Last widget selectable indices
        """
        return self.first_index, self.last_index

    def get_total_packed(self) -> int:
        """
        Return the total number of packed widgets.

        :return: Number of packed widgets
        """
        return len(self._widgets.values())

    def select(self, *args, **kwargs) -> 'Frame':
        return self

    def set_selection_effect(self, *args, **kwargs) -> 'Frame':
        pass

    def _apply_font(self) -> None:
        pass

    def _title_height(self) -> int:
        """
        Return the title height.

        :return: Title height in px
        """
        if not self._has_title or self._frame_title is None:
            return 0
        h = self._frame_title.get_height()
        h += self._frame_title.get_translate()[1]
        h += self._frame_title.get_attribute('pbottom')  # Bottom padding
        return h

    def _render(self) -> None:
        self._rect.height = self._frame_size[1] + self._title_height()
        self._rect.width = self._frame_size[0]

    def _draw(self, *args, **kwargs) -> None:
        pass

    def scale(self, *args, **kwargs) -> 'Frame':
        raise WidgetTransformationNotImplemented()

    def set_max_width(self, *args, **kwargs) -> 'Frame':
        raise WidgetTransformationNotImplemented()

    def set_max_height(self, *args, **kwargs) -> 'Frame':
        raise WidgetTransformationNotImplemented()

    def rotate(self, *args, **kwargs) -> 'Frame':
        raise WidgetTransformationNotImplemented()

    def flip(self, *args, **kwargs) -> 'Frame':
        raise WidgetTransformationNotImplemented()

    def get_decorator(self) -> 'Decorator':
        """
        Frame decorator belongs to the scrollarea decorator if enabled.

        :return: ScrollArea decorator
        """
        if self.is_scrollable:
            return self._frame_scrollarea.get_decorator()
        return self._decorator

    def get_index(self, widget: 'Widget') -> int:
        """
        Get index of the given widget within the widget list. Throws
        ``IndexError`` if widget does not exist.

        :param widget: Widget
        :return: Index
        """
        w = self.get_widgets(unpack_subframes=False)
        try:
            return w.index(widget)
        except ValueError:
            raise IndexError(f'{widget.get_class_id()} widget does not exist on {self.get_class_id()}')

    def set_position(self, x: NumberType, y: NumberType) -> 'Frame':
        if self._has_title:
            pad = self.get_padding()  # top, right, bottom, left
            tx, ty = self.get_translate()
            self._frame_title.set_position(x - pad[3] + tx, y - pad[0] + ty)
            for w in self._frame_title.get_widgets():
                w._set_position_relative_to_frame()
        super(Frame, self).set_position(x, y)
        if self.is_scrollable:
            self._frame_scrollarea.set_position(self._rect.x, self._rect.y + self._title_height())
        return self

    def draw(self, surface: 'pygame.Surface') -> 'Frame':
        if not self.is_visible():
            return self

        selected_widget: Optional['Widget'] = None

        # Simple case, no scrollarea
        if not self.is_scrollable:
            self.last_surface = surface
            self._draw_shadow(surface)
            self._draw_background_color(surface)
            self._decorator.draw_prev(surface)
            for widget in self._widgets.values():
                if widget.is_selected():
                    selected_widget = widget
                widget.draw(surface)
            if selected_widget is not None:
                selected_widget.draw_after_if_selected(surface)
            self._draw_border(surface)
            self._decorator.draw_post(surface)

        # Scrollarea
        else:
            self.last_surface = self._surface
            self._surface.fill((255, 255, 255, 0))
            self._draw_shadow(self._surface, rect=self._real_rect)
            self._draw_background_color(self._surface, rect=self._real_rect)
            scrollarea_decorator = self.get_decorator()
            scrollarea_decorator.force_cache_update()
            scrollarea_decorator.draw_prev(self._surface)
            for widget in self._widgets.values():
                if widget.is_selected():
                    selected_widget = widget
                widget.draw(self._surface)
            if selected_widget is not None:
                selected_widget.draw_after_if_selected(self._surface)
            self._frame_scrollarea.draw(surface)
            self._draw_border(surface)

        # If title
        if self._has_title:
            self._frame_title.draw(surface)

        self.apply_draw_callbacks()

        return self

    def _get_ht(self, widget: 'Widget', a: str) -> int:
        """
        Return the horizontal translation for widget.

        :param widget: Widget
        :param a: Horizontal alignment
        :return: Px
        """
        w = widget.get_width()
        # Some systems introduce 1px margin
        if w > (self._width + 1) and not self._relax:
            raise _FrameSizeException(
                f'{widget.get_class_id()} width ({w}) is greater than {self.get_class_id()}'
                f' width ({self._width}), try using widget.set_max_width(...) for '
                f'avoiding this issue, or set the widget as floating'
            )
        if a == ALIGN_CENTER:
            return int((self._width - w) / 2)
        elif a == ALIGN_RIGHT:
            return self._width - w
        else:  # Alignment left
            return 0

    def _get_vt(self, widget: 'Widget', v: str) -> int:
        """
        Return vertical translation for widget.

        :param widget: Widget
        :param v: Vertical position
        :return: Px
        """
        h = widget.get_height()
        if h > (self._height + 1) and not self._relax:
            raise _FrameSizeException(
                f'{widget.get_class_id()} height ({h}) is greater than {self.get_class_id()}'
                f' height ({self._height}), try using widget.set_max_height(...) '
                f'for avoiding this issue, or set the widget as floating'
            )
        if v == POSITION_CENTER:
            return int((self._height - h) / 2)
        elif v == POSITION_SOUTH:
            return self._height - h
        else:  # Position north
            return 0

    def _update_position_horizontal(self) -> None:
        """
        Compute widget position for horizontal orientation.
        """
        x_left = 0  # Total added to left
        x_right = 0  # Total added to right
        w_center = 0

        for w in self._widgets.values():
            align, v_pos = self._widgets_props[w.get_id()]
            if not w.is_visible(check_frame=False) or w.is_floating():
                continue
            if align == ALIGN_CENTER:
                w_center += w.get_width() + w.get_margin()[0]
                continue
            elif align == ALIGN_LEFT:
                x_left += w.get_margin()[0]
                self._pos[w.get_id()] = (x_left, self._get_vt(w, v_pos) + w.get_margin()[1])
                x_left += w.get_width()
            elif align == ALIGN_RIGHT:
                x_right -= (w.get_width() + w.get_margin()[0])
                self._pos[w.get_id()] = (self._width + x_right, self._get_vt(w, v_pos) + w.get_margin()[1])
            dw = x_left - x_right
            if dw > self._width and not self._relax:
                raise _FrameSizeException(
                    f'{w.get_class_id()} width ({dw}) exceeds {self.get_class_id()}'
                    f' width ({self._width}). Set frame._relax=True to ignore this Exception'
                )

        # Now center widgets
        available = self._width - (x_left - x_right)
        if w_center > available and not self._relax:
            raise _FrameSizeException(
                f'cannot place center widgets as required width ({w_center}) is '
                f'greater than available ({available}) in {self.get_class_id()}.'
                f' Set frame._relax=True to ignore this Exception'
            )
        x_center = int(self._width / 2 - w_center / 2)
        for w in self._widgets.values():
            align, v_pos = self._widgets_props[w.get_id()]
            if not w.is_visible(check_frame=False) or w.is_floating():
                continue
            if align == ALIGN_CENTER:
                x_center += w.get_margin()[0]
                self._pos[w.get_id()] = (x_center, self._get_vt(w, v_pos) + w.get_margin()[1])
                x_center += w.get_width()

    def _update_position_vertical(self) -> None:
        """
        Compute widget position for vertical orientation.
        """
        y_top = 0  # Total added to top
        y_bottom = 0  # Total added to bottom
        w_center = 0
        for w in self._widgets.values():
            align, v_pos = self._widgets_props[w.get_id()]
            if not w.is_visible(check_frame=False) or w.is_floating():
                continue
            if v_pos == POSITION_CENTER:
                w_center += w.get_height() + w.get_margin()[1]
                continue
            elif v_pos == POSITION_NORTH:
                y_top += w.get_margin()[1]
                self._pos[w.get_id()] = (self._get_ht(w, align) + w.get_margin()[0], y_top)
                y_top += w.get_height()
            elif v_pos == POSITION_SOUTH:
                y_bottom -= (w.get_height() + w.get_margin()[1])
                self._pos[w.get_id()] = (self._get_ht(w, align) + w.get_margin()[0], self._height + y_bottom)
            dh = y_top - y_bottom
            if dh > self._height and not self._relax:
                raise _FrameSizeException(
                    f'{w.get_class_id()} height ({dh}) exceeds {self.get_class_id()}'
                    f' height ({self._height}). Set frame._relax=True to ignore '
                    f'this Exception'
                )

        # Now center widgets
        available = self._height - (y_top - y_bottom)
        if w_center > available and not self._relax:
            raise _FrameSizeException(
                f'cannot place center widgets as required height ({w_center}) is'
                f' greater than available ({available}) in {self.get_class_id()}.'
                f' Set frame._relax=True to ignore this Exception'
            )
        y_center = int(self._height / 2 - w_center / 2)
        for w in self._widgets.values():
            align, v_pos = self._widgets_props[w.get_id()]
            if not w.is_visible(check_frame=False) or w.is_floating():
                continue
            if v_pos == POSITION_CENTER:
                y_center += w.get_margin()[1]
                self._pos[w.get_id()] = (self._get_ht(w, align) + w.get_margin()[0], y_center)
                y_center += w.get_height()

    def update_position(self) -> 'Frame':
        """
        Update the position of each widget.

        :return: Self reference
        """
        if len(self._widgets) == 0:
            return self

        # Update position based on orientation
        if self._orientation == ORIENTATION_HORIZONTAL:
            self._update_position_horizontal()
        elif self._orientation == ORIENTATION_VERTICAL:
            self._update_position_vertical()

        # Apply position to each widget
        for w in self._widgets.keys():
            widget = self._widgets[w]
            if not widget.is_visible(check_frame=False):
                widget.set_position(0, 0)
                continue
            if widget.is_floating():
                tx = 0
                ty = 0
            else:
                tx, ty = self._pos[w]
            ty += self._title_height()
            if widget.get_menu() is None:  # Widget is only appended to Frame
                fx, fy = self.get_position()
                margin = widget.get_margin()
                padding = widget.get_padding()
                widget.set_position(fx + margin[0] + padding[3], fy + padding[0])
            if self.is_scrollable and isinstance(widget, Frame) and widget.is_scrollable:
                widget.get_scrollarea(inner=True).set_position(tx, ty)
            # If scrollarea, subtract this position to each widget
            if self._frame_scrollarea is not None:
                sx, sy = self._frame_scrollarea.get_position()
                tx -= sx
                ty -= sy
            widget._translate_virtual = (tx, ty)  # Translate to scrollarea

        # Check if control widget has changed positioning. This fixes centering issues
        if self._control_widget is not None:
            c_pos = self._control_widget.get_position()
            if self._control_widget_last_pos != c_pos:
                self._control_widget_last_pos = c_pos
                if self._recursive_render <= 100 and self._menu is not None:
                    self._menu.render()
                self._recursive_render += 1
            else:
                self._recursive_render = 0

        # If frame has title
        if self._has_title:
            self._frame_title.update_position()

        return self

    def get_widgets(
        self,
        unpack_subframes: bool = True,
        unpack_subframes_include_frame: bool = False,
        reverse: bool = False
    ) -> Tuple['Widget', ...]:
        """
        Get widgets as a tuple.

        :param unpack_subframes: If ``True`` add frame widgets instead of frame
        :param unpack_subframes_include_frame: If ``True`` the unpacked frame is also added to the widget list
        :param reverse: Reverse the returned tuple
        :return: Widget tuple
        """
        wtp = []
        for widget in self._widgets.values():
            if isinstance(widget, Frame) and unpack_subframes:
                if unpack_subframes_include_frame:
                    wtp.append(widget)
                ww = widget.get_widgets(
                    unpack_subframes=unpack_subframes,
                    unpack_subframes_include_frame=unpack_subframes_include_frame
                )
                for i in ww:
                    wtp.append(i)
                continue
            wtp.append(widget)
        if reverse:
            wtp.reverse()
        return tuple(wtp)

    def clear(self) -> Union['Widget', Tuple['Widget', ...]]:
        """
        Unpack all widgets within frame.

        :return: Removed widgets
        """
        unpacked = []
        for w in self.get_widgets(unpack_subframes=False):
            self.unpack(w)
            unpacked.append(w)
        return tuple(unpacked)

    def resize(
        self,
        width: NumberType,
        height: NumberType,
        max_width: Optional[NumberType] = None,
        max_height: Optional[NumberType] = None
    ) -> 'Frame':
        """
        Resize the Frame.

        :param width: New width in px. Horizontal padding will be subtracted
        :param height: New height in px. Vertical padding will be subtracted
        :param max_width: Max frame width if the Frame is scrollable. If ``None`` the same width will be used
        :param max_height: Max frame height if the Frame is scrollable. If ``None`` the same height will be used
        :return: Self reference
        """
        assert isinstance(width, NumberInstance)
        assert isinstance(height, NumberInstance)

        pad_h = self._padding[1] + self._padding[3]
        pad_v = self._padding[0] + self._padding[2]

        # Subtract padding
        width -= pad_h
        height -= pad_v

        # Check size
        assert width > 0 and height > 0, 'new width and height must be greater than zero'

        # Update width/height
        if width < self._width or height < self._height:
            self.relax()
        self._frame_size = (width, height)  # Size of the frame, set in make_scrollarea
        self._height = int(height)
        self._real_rect = pygame.Rect(0, 0, width, height)
        self._width = int(width)

        # Get previous buttons if it has title
        prev_has_title = self._has_title
        prev_title_frame = self._frame_title
        prev_title_buttons = list(self._frame_title.get_widgets()) if self._has_title else []
        if len(prev_title_buttons) >= 1:  # Pop label
            prev_title_buttons.pop(0)

        # Make scrollable if scrollable
        if self.is_scrollable:
            assert self._frame_scrollarea.has_attribute('constructor'), \
                'frame scrollarea does not have the "constructor" attribute. Make ' \
                'sure the scrollarea has been created using make_scrollarea() method'
            kwargs: Dict[str, Any] = self._frame_scrollarea.get_attribute('constructor')
            if max_width is None:
                max_width = width
            if max_height is None:
                max_height = height
            self.make_scrollarea(
                max_height=max_height,
                max_width=max_width,
                scrollarea_color=kwargs['scrollarea_color'],
                scrollbar_color=kwargs['scrollbar_color'],
                scrollbar_cursor=kwargs['scrollbar_cursor'],
                scrollbar_shadow=kwargs['scrollbar_shadow'],
                scrollbar_shadow_color=kwargs['scrollbar_shadow_color'],
                scrollbar_shadow_offset=kwargs['scrollbar_shadow_offset'],
                scrollbar_shadow_position=kwargs['scrollbar_shadow_position'],
                scrollbar_slider_color=kwargs['scrollbar_slider_color'],
                scrollbar_slider_hover_color=kwargs['scrollbar_slider_hover_color'],
                scrollbar_slider_pad=kwargs['scrollbar_slider_pad'],
                scrollbar_thick=kwargs['scrollbar_thick'],
                scrollbars=kwargs['scrollbars']
            )
        else:
            assert max_width is None and max_height is None, \
                'if previous Frame is not scrollable (make_scrollarea has been ' \
                'called) max_width and max_height must be None'

        # If had title, remove and create a new one
        if self._has_title:
            self.remove_title()
        if prev_has_title:
            for btn in prev_title_buttons:
                prev_title_frame.unpack(btn)
                btn.set_margin(0, 0)
                btn.set_float(False)
            assert prev_title_frame.has_attribute('constructor'), \
                'frame title does not have the attribute "constructor". Make sure ' \
                'the frame title has been created through set_title() method.'
            kwargs = prev_title_frame.get_attribute('constructor')
            new_title_frame = self.set_title(
                title=kwargs['title'],
                cursor=kwargs['cursor'],
                background_color=kwargs['background_color'],
                draggable=kwargs['draggable'],
                padding_inner=kwargs['padding_inner'],
                padding_outer=kwargs['padding_outer'],
                title_alignment=kwargs['title_alignment'],
                title_buttons_alignment=kwargs['title_buttons_alignment'],
                title_font=kwargs['title_font'],
                title_font_color=kwargs['title_font_color'],
                title_font_size=kwargs['title_font_size']
            )

            # Pack previous buttons
            # prev_title_buttons.reverse()
            for btn in prev_title_buttons:
                align = btn.get_attribute('align', kwargs['title_buttons_alignment'])
                margin = btn.get_attribute('margin', (0, 0))
                new_title_frame.pack(btn, align=align, margin=margin)

        # Force render
        self._render()
        self.force_menu_surface_update()

        return self

    def unfloat(self) -> 'Frame':
        """
        Disable float status for each subwidget.

        :return: Self reference
        """
        for w in self.get_widgets(unpack_subframes_include_frame=True):
            w.set_float(False)
        if self._menu is not None:
            self._menu.render()
            self._menu.scroll_to_widget(None)
        return self

    def get_scrollarea(self, inner: bool = False) -> Optional['pygame_menu._scrollarea.ScrollArea']:
        """
        Return the scrollarea object.

        :param inner: If ``True`` return the inner scrollarea
        :return: ScrollArea object
        """
        if inner:
            return self._frame_scrollarea
        return self._scrollarea

    def set_frame(self, frame: 'pygame_menu.widgets.Frame') -> 'Frame':
        assert self != frame, \
            f'{frame.get_class_id()} cannot set itself as a frame'
        super(Frame, self).set_frame(frame)
        if self._frame_title is not None:
            self._frame_title.set_frame(frame)
        return self

    def set_scrollarea(self, scrollarea: Optional['pygame_menu._scrollarea.ScrollArea']) -> None:
        if scrollarea is not None:
            assert scrollarea != self._frame_scrollarea, \
                f'scrollarea cannot be {self.get_class_id()}._frame_scrollarea {scrollarea.get_class_id()}'
        self._scrollarea = scrollarea
        if self._frame_scrollarea is not None:
            self._frame_scrollarea.set_parent_scrollarea(scrollarea)
        else:
            for w in self.get_widgets(unpack_subframes=False):
                w.set_scrollarea(scrollarea)
        if self._frame_title is not None:
            self._frame_title.set_scrollarea(scrollarea)

    def scrollh(self, value: NumberType) -> 'Frame':
        """
        Scroll to horizontal value if frame is scrollable.

        :param value: Horizontal scroll value, if ``0`` scroll to left; ``1`` scroll to right
        :return: Self reference
        """
        if self._frame_scrollarea is not None:
            self._frame_scrollarea.scroll_to(ORIENTATION_HORIZONTAL, value)
        return self

    def scrollv(self, value: NumberType) -> 'Frame':
        """
        Scroll to vertical value if frame is scrollable.

        :param value: Vertical scroll value, if ``0`` scroll to top; ``1`` scroll to bottom
        :return: Self reference
        """
        if self._frame_scrollarea is not None:
            self._frame_scrollarea.scroll_to(ORIENTATION_VERTICAL, value)
        return self

    def get_scroll_value_percentage(self, orientation: str) -> float:
        """
        Get the scroll value in percentage, if ``0`` the scroll is at top/left,
        ``1`` bottom/right.

        .. note::

            If ScrollArea does not contain such orientation scroll, or frame is not scrollable,
            ``-1`` is returned.

        :param orientation: Orientation. See :py:mod:`pygame_menu.locals`
        :return: Value from ``0`` to ``1``
        """
        if self._frame_scrollarea is not None:
            return self._frame_scrollarea.get_scroll_value_percentage(orientation)
        return -1

    def unpack(self, widget: 'Widget') -> 'Frame':
        """
        Unpack widget from Frame. If widget does not exist, raises ``ValueError``.
        Unpacked widgets adopt a floating position and are moved to the last position
        of the widget list of Menu

        :param widget: Widget to unpack
        :return: Unpacked widget
        """
        assert widget != self, 'frame cannot unpack itself'
        assert len(self._widgets) > 0, 'frame is empty'
        wid = widget.get_id()
        if wid not in self._widgets.keys():
            raise ValueError(f'{widget.get_class_id()} does not exist in {self.get_class_id()}')
        assert widget._frame == self, 'widget frame differs from current'
        widget.set_float()
        if self._menu is not None:
            widget.set_scrollarea(self._menu.get_scrollarea())
        widget._frame = None
        widget._translate_virtual = (0, 0)
        del self._widgets[wid]
        try:
            del self._pos[wid]
        except KeyError:
            pass

        # Move widget to the last position of widget list
        menu_widgets = self._get_menu_widgets()
        if widget.get_menu() == self._menu and widget in menu_widgets:
            self._menu._validate_frame_widgetmove = False
            try:
                self._menu.move_widget_index(widget, render=False)
            # Assertion error if moving widget (last) to same position (last)
            except (ValueError, AssertionError):
                pass
            if isinstance(widget, Frame):
                widgets = widget.get_widgets(unpack_subframes_include_frame=True)
                for w in widgets:
                    if w.get_menu() is None or w not in menu_widgets:
                        continue
                    self._menu.move_widget_index(w, render=False)
            self._menu._validate_frame_widgetmove = True

        # Check if frame contains more frames
        self._has_frames = False
        for k in self.get_widgets(unpack_subframes=False):
            if isinstance(k, Frame):
                self._has_frames = True
                break

        # Update selected
        if self._menu is not None:
            self._menu.move_widget_index(None, update_selected_index=True)

        # Update indices
        self.update_indices()

        # Render menu
        self._menu_render()

        if self._control_widget == widget:
            self._control_widget = None
            self._control_widget_last_pos = None
            for w in self.get_widgets():
                if w.get_menu() is not None:
                    self._control_widget = w
                    self._control_widget_last_pos = self._control_widget.get_position()
                    break

        if widget.is_selected():
            widget.scroll_to_widget()

        if len(self._widgets) == 0:  # Scroll to top
            self.scrollv(0)
            self.scrollh(0)

        if isinstance(widget, Frame):
            self._sort_menu_update_frames()

        # Update widget leave
        check_widget_mouseleave()

        return widget

    def pack(
        self,
        widget: Union['Widget', List['Widget'], Tuple['Widget', ...]],
        align: str = ALIGN_LEFT,
        vertical_position: str = POSITION_NORTH,
        margin: Vector2NumberType = (0, 0)
    ) -> Union['Widget', List['Widget'], Tuple['Widget', ...], Any]:
        """
        Packs widget in the frame line. To pack a widget it has to be already
        appended to Menu, and the Menu must be the same as the frame.

        Packing is added to the same line, for example if three LEFT widgets are
        added:

            .. code-block:: python

                <frame horizontal>
                frame.pack(W1, alignment=ALIGN_LEFT, vertical_position=POSITION_NORTH)
                frame.pack(W2, alignment=ALIGN_LEFT, vertical_position=POSITION_CENTER)
                frame.pack(W3, alignment=ALIGN_LEFT, vertical_position=POSITION_SOUTH)

                ----------------
                |W1            |
                |   W2         |
                |      W3      |
                ----------------

        Another example:

            .. code-block:: python

                <frame horizontal>
                frame.pack(W1, alignment=ALIGN_LEFT)
                frame.pack(W2, alignment=ALIGN_CENTER)
                frame.pack(W3, alignment=ALIGN_RIGHT)

                ----------------
                |W1    W2    W3|
                ----------------

            .. code-block:: python

                <frame vertical>
                frame.pack(W1, alignment=ALIGN_LEFT)
                frame.pack(W2, alignment=ALIGN_CENTER)
                frame.pack(W3, alignment=ALIGN_RIGHT)

                --------
                |W1    |
                |  W2  |
                |    W3|
                --------

        .. note::

            Frame does not consider previous widget margin. For such purpose, use
            ``margin`` pack parameter.

        .. note::

            It is recommended to force menu rendering after packing all widgets.

        .. note::

            Packing applies a virtual translation to the widget, previous translation
            is not modified.

        .. note::

            Widget floating is also considered within frames. If a widget is floating,
            it does not add any size to the respective positioning.

        :param widget: Widget to be packed
        :param align: Widget alignment. See :py:mod:`pygame_menu.locals`
        :param vertical_position: Vertical position of the widget within frame. Only valid: north, center, and south. See :py:mod:`pygame_menu.locals`
        :param margin: (left, top) margin of added widget in px. It overrides the previous widget margin
        :return: Added widget references
        """
        assert self._menu is not None or self._menu_can_be_none_pack, \
            f'{self.get_class_id()} menu must be set before packing widgets'
        if isinstance(widget, VectorInstance):
            for w in widget:
                self.pack(widget=w, align=align, vertical_position=vertical_position)
            return widget
        assert isinstance(widget, Widget)
        if isinstance(widget, Frame):
            assert widget.get_menu() is not None or self._menu_can_be_none_pack, \
                f'{widget.get_class_id()} menu cannot be None'
        assert widget.get_id() not in self._widgets.keys(), \
            f'{widget.get_class_id()} already exists in {self.get_class_id()}'
        assert widget.get_menu() == self._menu or widget.get_menu() is None, \
            'widget menu to be added to frame must be in same menu as frame, or ' \
            'it can have any Menu instance'
        assert widget.get_frame() is None, \
            f'{widget.get_class_id()} is already packed in {widget.get_frame().get_class_id()}'
        assert_alignment(align)
        assert vertical_position in (POSITION_NORTH, POSITION_CENTER, POSITION_SOUTH), \
            'vertical position must be NORTH, CENTER, or SOUTH'
        assert_vector(margin, 2)
        assert widget.configured, \
            f'{widget.get_class_id()} must be configured before packing'

        if widget.get_margin() != (0, 0) and self._pack_margin_warning:
            if self._verbose:
                warn(
                    f'{widget.get_class_id()} margin should be (0, 0) if packed, but'
                    f' received {widget.get_margin()}; {self.get_class_id()}.pack() '
                    f'does not consider previous widget margin. Set '
                    f'frame._pack_margin_warning=False to hide this warning'
                )

        if isinstance(widget, Frame):
            widget.update_indices()

        widget.set_frame(self)
        widget.set_margin(*margin)
        if self._frame_scrollarea is not None:
            widget.set_scrollarea(self._frame_scrollarea)
        else:
            widget.set_scrollarea(self._scrollarea)
        if self.is_scrollable or self._has_title or isinstance(widget, Frame):
            self._sort_menu_update_frames()
        self._widgets[widget.get_id()] = widget
        self._widgets_props[widget.get_id()] = (align, vertical_position)

        # Sort widgets to keep selection order
        menu_widgets = self._get_menu_widgets()
        if widget.get_menu() is not None and widget in menu_widgets:
            self._menu._validate_frame_widgetmove = False
            widgets_list = list(self._widgets.values())

            # Move frame to last
            if len(self._widgets) > 1:
                w_last = widgets_list[-2]  # -1 is the last added
                for i in range(2, len(self._widgets)):
                    if w_last.get_menu() is None and len(self._widgets) > 2:
                        w_last = widgets_list[-(i + 1)]
                    else:
                        break

                # Check for last if w_last is frame
                while True:
                    if not (isinstance(w_last, Frame) and w_last.get_indices() != (-1, -1)) or w_last.get_menu() is None:
                        break
                    w_last = menu_widgets[w_last.last_index]

                if w_last.get_menu() == self._menu:
                    self._menu.move_widget_index(self, w_last, render=False)

            # Swap
            self._menu.move_widget_index(widget, self, render=False)
            if isinstance(widget, Frame):
                reverse = menu_widgets.index(widget) == len(menu_widgets) - 1
                widgs = widget.get_widgets(unpack_subframes_include_frame=True, reverse=reverse)

                first_moved_widget = None
                last_moved_widget = None
                for w in widgs:
                    if w.get_menu() is None or w not in menu_widgets:
                        continue
                    self._menu.move_widget_index(w, self, render=False)
                    if first_moved_widget is None:
                        first_moved_widget = w
                    last_moved_widget = w

                swap_target = last_moved_widget if reverse else first_moved_widget
                if swap_target is not None:
                    menu_widgets.remove(widget)
                    menu_widgets.insert(menu_widgets.index(swap_target), widget)

            # Move widget to first
            menu_widgets.remove(self)
            for k in range(len(widgets_list)):
                if widgets_list[k].get_menu() == self._menu:
                    menu_widgets.insert(menu_widgets.index(widgets_list[k]), self)
                    break
            self._menu._validate_frame_widgetmove = True

            # Update control widget
            if self._control_widget is None:
                self._control_widget = widget
                self._control_widget_last_pos = self._control_widget.get_position()

        if isinstance(widget, Frame):
            self._has_frames = True

        # Update menu selected widget
        if self._menu is not None:
            self._menu.move_widget_index(None, update_selected_index=True)

        # Render is mandatory as it modifies row/column layout
        try:
            self.update_position()
            self._menu_render()
        except _FrameSizeException:
            self.unpack(widget)
            raise

        # Request scroll if widget is selected
        if widget.is_selected():
            widget.scroll_to_widget()
            widget.scroll_to_widget()

        # Update widget leave
        check_widget_mouseleave()

        return widget

    def contains_widget(self, widget: 'Widget') -> bool:
        """
        Return true if the frame contains the given widget.

        :param widget: Widget to check
        :return: ``True`` if widget within frame
        """
        return widget.get_frame() == self and widget.get_id() in self._widgets.keys()

    def hide(self) -> 'Frame':
        super(Frame, self).hide()
        if self._has_title:
            self._frame_title.hide()
        # sub-widgets cannot be hidden because some widgets compute sizing even
        # if the frame itself is hidden
        # for w in self.get_widgets(unpack_subframes=False):
        #     w.hide()
        return self

    def show(self) -> 'Frame':
        super(Frame, self).show()
        # same as hiding, sub-widgets should not be modified
        # for w in self.get_widgets(unpack_subframes=False):
        #     w.show()
        if self._has_title:
            self._frame_title.show()
        return self

    def update_indices(self) -> 'Frame':
        """
        Update first and last selectable widget index.

        :return: Self reference
        """
        # Public index update only triggered if frame does not contain subframes
        if not self._has_frames:
            self._update_indices()
        return self

    def _update_indices(self) -> None:
        """
        Private update indices method.
        """
        self.first_index = -1
        self.last_index = -1
        for widget in self.get_widgets(unpack_subframes=False):
            if (widget.is_selectable or isinstance(widget, Frame)) and widget.get_menu() is not None:
                if isinstance(widget, Frame) and widget.get_indices() == (-1, -1):
                    continue  # Frames with not selectable indices are not counted
                windex = widget.get_col_row_index()[2]
                if self.first_index == -1:
                    self.first_index = windex
                    self.last_index = windex
                else:
                    self.first_index = min(self.first_index, windex)
                    self.last_index = max(self.last_index, windex)
        if self.get_frame() is not None:
            self.get_frame()._update_indices()

    def update(self, events: EventVectorType) -> bool:
        self.apply_update_callbacks(events)

        updated = False
        if self.readonly or not self.is_visible():
            self._readonly_check_mouseover(events)
            return updated

        # Check title events
        if self._has_title:
            # Check for buttons
            for w in self._frame_title.get_widgets():
                updated = updated or w.update(events)

            # Check if clicked the title for drag
            for event in events:

                # Check mouseover
                self._frame_title._check_mouseover(event)

                # If clicked in title
                if self._draggable and (
                    event.type == pygame.MOUSEBUTTONDOWN and self._mouse_enabled and event.button in (1, 2, 3) or
                    event.type == FINGERDOWN and self._touchscreen_enabled and self._menu is not None
                ):
                    event_pos = get_finger_pos(self._menu, event)

                    if self._frame_title.get_rect(to_real_position=True).collidepoint(*event_pos):
                        if not self._frame_title.get_attribute('drag', False):
                            self._frame_title.set_attribute('drag', True)
                            updated = True

                # User releases the button
                elif (
                    event.type == pygame.MOUSEBUTTONUP and self._mouse_enabled and event.button in (1, 2, 3) or
                    event.type == FINGERUP and self._touchscreen_enabled
                ):
                    self._frame_title.set_attribute('drag', False)

                # Mouse out from window
                # elif event.type == pygame.ACTIVEEVENT:
                #     if event.gain != 1:
                #         self._frame_title.set_attribute('drag', False)
                #         break

                # User moves the mouse while drag
                elif (
                    event.type == pygame.MOUSEMOTION and hasattr(event, 'rel') or
                    event.type == FINGERMOTION and self._touchscreen_enabled and self._menu is not None
                ):

                    if self._frame_title.get_attribute('drag', False) and self._draggable:
                        # Get relative movement
                        rx = event.rel[0] if event.type == pygame.MOUSEMOTION else \
                            event.dx * self._menu.get_window_size()[0] * S_FINGER_FACTOR[0]
                        ry = event.rel[1] if event.type == pygame.MOUSEMOTION else \
                            event.dy * self._menu.get_window_size()[1] * S_FINGER_FACTOR[1]

                        event_pos = get_finger_pos(self._menu, event)
                        tx, ty = self.get_translate()
                        title_rect = self._frame_title.get_rect(to_real_position=True)

                        if self._rect.y <= 0:
                            if not title_rect.collidepoint(*event_pos):
                                if ry > 0:
                                    self.translate(tx, ty - self._rect.y)
                                    self.force_menu_surface_update()
                                    updated = True
                                continue

                        elif self.get_scrollarea() is not None:
                            max_v = self.get_scrollarea().get_world_size()[1] - self._title_height()
                            if self._rect.y >= max_v:
                                if not title_rect.collidepoint(*event_pos):
                                    if ry < 0:
                                        continue
                                if ry > 0:
                                    continue
                            if ry > 0 and self._rect.y + ry >= max_v:
                                continue

                        # Get the max/min distance which can translate in vertical
                        if ry < 0:
                            ry = -min(-ry, self._rect.y)

                        self.translate(tx + rx, ty + ry)
                        self.force_menu_surface_update()
                        updated = True

        # Check mouseover
        for event in events:
            if self._check_mouseover(event):
                break

        # If not scrollable, return
        if not self.is_scrollable:
            return updated

        return updated or self._frame_scrollarea.update(events)


class _FrameSizeException(Exception):
    """
    If widget size is greater than frame raises exception.
    """
    pass


class _FrameDoNotAcceptScrollarea(Exception):
    """
    Raised if the frame does not accept a scrollarea.
    """
    pass


class _FrameDoNotAcceptTitle(Exception):
    """
    Raised if the frame does not accept a title.
    """
    pass


class FrameManager(AbstractWidgetManager, ABC):
    """
    Frame manager.
    """

    def _frame(
        self,
        width: NumberType,
        height: NumberType,
        orientation: str,
        frame_id: str = '',
        **kwargs
    ) -> 'pygame_menu.widgets.Frame':
        """
        Adds a frame to the Menu.

        :param width: Frame width in px
        :param height: Frame height in px
        :param orientation: Frame orientation, horizontal or vertical. See :py:mod:`pygame_menu.locals`
        :param frame_id: ID of the frame
        :param kwargs: Optional keyword arguments
        :return: Widget object
        :rtype: :py:class:`pygame_menu.widgets.Frame`
        """
        from pygame_menu._scrollarea import get_scrollbars_from_position

        # Remove invalid keys from kwargs
        for key in list(kwargs.keys()):
            if key not in ('align', 'background_color', 'background_inflate',
                           'border_color', 'border_inflate', 'border_width',
                           'cursor', 'margin', 'padding', 'max_height', 'max_width',
                           'scrollbar_color', 'scrollbar_cursor',
                           'scrollbar_shadow_color', 'scrollbar_shadow_offset',
                           'scrollbar_shadow_position', 'scrollbar_shadow',
                           'scrollbar_slider_color', 'scrollbar_slider_pad',
                           'scrollbar_thick', 'scrollbars', 'scrollarea_color',
                           'border_position', 'scrollbar_slider_hover_color',
                           'tab_size', 'float', 'float_origin_position'):
                kwargs.pop(key, None)

        attributes = self._filter_widget_attributes(kwargs)
        pad = parse_padding(attributes['padding'])  # top, right, bottom, left
        pad_h = pad[1] + pad[3]
        pad_v = pad[0] + pad[2]

        assert width > pad_h, \
            f'frame width ({width}) cannot be lower than horizontal padding size ({pad_h})'
        assert height > pad_v, \
            f'frame height ({height}) cannot be lower than vertical padding size ({pad_v})'

        widget = Frame(
            width=width - pad_h,
            height=height - pad_v,
            orientation=orientation,
            frame_id=frame_id
        )
        self._configure_widget(widget=widget, **attributes)

        widget.make_scrollarea(
            max_height=kwargs.pop('max_height', height) - pad_v,
            max_width=kwargs.pop('max_width', width) - pad_h,
            scrollarea_color=kwargs.pop('scrollarea_color', None),
            scrollbar_color=kwargs.pop('scrollbar_color', self._theme.scrollbar_color),
            scrollbar_cursor=kwargs.pop('scrollbar_cursor', self._theme.scrollbar_cursor),
            scrollbar_shadow=kwargs.pop('scrollbar_shadow', self._theme.scrollbar_shadow),
            scrollbar_shadow_color=kwargs.pop('scrollbar_shadow_color', self._theme.scrollbar_shadow_color),
            scrollbar_shadow_offset=kwargs.pop('scrollbar_shadow_offset', self._theme.scrollbar_shadow_offset),
            scrollbar_shadow_position=kwargs.pop('scrollbar_shadow_position', self._theme.scrollbar_shadow_position),
            scrollbar_slider_color=kwargs.pop('scrollbar_slider_color', self._theme.scrollbar_slider_color),
            scrollbar_slider_hover_color=kwargs.pop('scrollbar_slider_hover_color', self._theme.scrollbar_slider_hover_color),
            scrollbar_slider_pad=kwargs.pop('scrollbar_slider_pad', self._theme.scrollbar_slider_pad),
            scrollbar_thick=kwargs.pop('scrollbar_thick', self._theme.scrollbar_thick),
            scrollbars=get_scrollbars_from_position(kwargs.pop('scrollbars', self._theme.scrollarea_position))
        )

        self._append_widget(widget)
        self._check_kwargs(kwargs)

        return widget

    def frame_h(
        self,
        width: NumberType,
        height: NumberType,
        frame_id: str = '',
        **kwargs
    ) -> 'pygame_menu.widgets.Frame':
        """
        Adds a horizontal frame to the Menu. Frame is a widget container that
        packs many widgets within. All contained widgets have a floating position,
        and use only 1 position in column/row layout.

        .. code-block:: python

            frame.pack(W1, alignment=ALIGN_LEFT, vertical_position=POSITION_NORTH)
            frame.pack(W2, alignment=ALIGN_LEFT, vertical_position=POSITION_CENTER)
            frame.pack(W3, alignment=ALIGN_LEFT, vertical_position=POSITION_SOUTH)
            ...

            ----------------
            |W1            |
            |   W2     ... |
            |      W3      |
            ----------------

        kwargs (Optional)
            - ``align``                         (str) – Widget `alignment <https://pygame-menu.readthedocs.io/en/latest/_source/themes.html#alignment>`_
            - ``background_color``              (tuple, list, str, int, :py:class:`pygame.Color`, :py:class:`pygame_menu.baseimage.BaseImage`) – Color of the background. ``None`` for no-color
            - ``background_inflate``            (tuple, list) – Inflate background on x-axis and y-axis (x, y) in px
            - ``border_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Widget border color. ``None`` for no-color
            - ``border_inflate``                (tuple, list) – Widget border inflate on x-axis and y-axis (x, y) in px
            - ``border_position``               (str, tuple, list) – Widget border positioning. It can be a single position, or a tuple/list of positions. Only are accepted: north, south, east, and west. See :py:mod:`pygame_menu.locals`
            - ``border_width``                  (int) – Border width in px. If ``0`` disables the border
            - ``cursor``                        (int, :py:class:`pygame.cursors.Cursor`, None) – Cursor of the frame if the mouse is placed over
            - ``float``                         (bool) - If ``True`` the widget don't contribute width/height to the Menu widget positioning computation, and don't add one unit to the rows
            - ``float_origin_position``         (bool) - If ``True`` the widget position is set to the top-left position of the Menu if the widget is floating
            - ``margin``                        (tuple, list) – Widget (left, bottom) margin in px
            - ``max_height``                    (int) – Max height in px. If this value is lower than the frame ``height`` a scrollbar will appear on vertical axis. ``None`` by default (same height)
            - ``max_width``                     (int) – Max width in px. If this value is lower than the frame ``width`` a scrollbar will appear on horizontal axis. ``None`` by default (same width)
            - ``padding``                       (int, float, tuple, list) – Widget padding according to CSS rules. General shape: (top, right, bottom, left)
            - ``scrollarea_color``              (tuple, list, str, int, :py:class:`pygame.Color`, :py:class:`pygame_menu.baseimage.BaseImage`,None) – Scroll area color. If ``None`` area is transparent
            - ``scrollbar_color``               (tuple, list, str, int, :py:class:`pygame.Color`) – Scrollbar color
            - ``scrollbar_cursor``              (int, :py:class:`pygame.cursors.Cursor`, None) – Cursor of the scrollbars if the mouse is placed over
            - ``scrollbar_shadow_color``        (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the shadow of each scrollbar
            - ``scrollbar_shadow_offset``       (int) – Offset of the scrollbar shadow in px
            - ``scrollbar_shadow_position``     (str) – Position of the scrollbar shadow. See :py:mod:`pygame_menu.locals`
            - ``scrollbar_shadow``              (bool) – Indicate if a shadow is drawn on each scrollbar
            - ``scrollbar_slider_color``        (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the sliders
            - ``scrollbar_slider_hover_color``  (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the slider if hovered or clicked
            - ``scrollbar_slider_pad``          (int, float) – Space between slider and scrollbars borders in px
            - ``scrollbar_thick``               (int) – Scrollbar thickness in px
            - ``scrollbars``                    (str) – Scrollbar position. See :py:mod:`pygame_menu.locals`
            - ``shadow_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the widget shadow
            - ``shadow_radius``                 (int) - Border radius of the shadow
            - ``shadow_type``                   (str) - Shadow type, it can be ``'rectangular'`` or ``'ellipse'``
            - ``shadow_width``                  (int) - Width of the shadow. If ``0`` the shadow is disabled

        .. note::

            All theme-related optional kwargs use the default Menu theme if not
            defined.

        .. note::

            If horizontal frame contains a scrollarea (setting ``max_height`` or
            ``max_width`` less than size) padding will be set at zero.

        .. note::

            Packing applies a virtual translation to the widget, previous translation
            is not modified.

        .. note::

            Widget floating is also considered within frames. If a widget is
            floating, it does not add any size to the respective positioning.

        .. note::

            The Frame size created with this method does consider the padding. Thus,
            if Frame is created with ``width=100``, ``height=200`` and ``padding=25``
            the final internal size is ``width=50`` and ``height=150``.

        .. note::

            This is applied only to the base Menu (not the currently displayed,
            stored in ``_current`` pointer); for such behaviour apply to
            :py:meth:`pygame_menu.menu.Menu.get_current` object.

        :param width: Frame width in px
        :param height: Frame height in px
        :param frame_id: ID of the horizontal frame
        :param kwargs: Optional keyword arguments
        :return: Widget object
        :rtype: :py:class:`pygame_menu.widgets.Frame`
        """
        return self._frame(width, height, ORIENTATION_HORIZONTAL, frame_id, **kwargs)

    def frame_v(
        self,
        width: NumberType,
        height: NumberType,
        frame_id: str = '',
        **kwargs
    ) -> 'pygame_menu.widgets.Frame':
        """
        Adds a vertical frame to the Menu. Frame is a widget container that packs
        many widgets within. All contained widgets have a floating position, and
        use only 1 position in column/row layout.

        .. code-block:: python

            frame.pack(W1, alignment=ALIGN_LEFT)
            frame.pack(W2, alignment=ALIGN_CENTER)
            frame.pack(W3, alignment=ALIGN_RIGHT)
            ...

            --------
            |W1    |
            |  W2  |
            |    W3|
            | ...  |
            --------

        kwargs (Optional)
            - ``align``                         (str) – Widget `alignment <https://pygame-menu.readthedocs.io/en/latest/_source/themes.html#alignment>`_
            - ``background_color``              (tuple, list, str, int, :py:class:`pygame.Color`, :py:class:`pygame_menu.baseimage.BaseImage`) – Color of the background. ``None`` for no-color
            - ``background_inflate``            (tuple, list) – Inflate background on x-axis and y-axis (x, y) in px
            - ``border_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Widget border color. ``None`` for no-color
            - ``border_inflate``                (tuple, list) – Widget border inflate on x-axis and y-axis (x, y) in px
            - ``border_position``               (str, tuple, list) – Widget border positioning. It can be a single position, or a tuple/list of positions. Only are accepted: north, south, east, and west. See :py:mod:`pygame_menu.locals`
            - ``border_width``                  (int) – Border width in px. If ``0`` disables the border
            - ``cursor``                        (int, :py:class:`pygame.cursors.Cursor`, None) – Cursor of the frame if the mouse is placed over
            - ``float``                         (bool) - If ``True`` the widget don't contribute width/height to the Menu widget positioning computation, and don't add one unit to the rows
            - ``float_origin_position``         (bool) - If ``True`` the widget position is set to the top-left position of the Menu if the widget is floating
            - ``margin``                        (tuple, list) – Widget (left, bottom) margin in px
            - ``max_height``                    (int) – Max height in px. If this value is lower than the frame ``height`` a scrollbar will appear on vertical axis. ``None`` by default (same height)
            - ``max_width``                     (int) – Max width in px. If this value is lower than the frame ``width`` a scrollbar will appear on horizontal axis. ``None`` by default (same width)
            - ``padding``                       (int, float, tuple, list) – Widget padding according to CSS rules. General shape: (top, right, bottom, left)
            - ``scrollarea_color``              (tuple, list, str, int, :py:class:`pygame.Color`, :py:class:`pygame_menu.baseimage.BaseImage`,None) – Scroll area color. If ``None`` area is transparent
            - ``scrollbar_color``               (tuple, list, str, int, :py:class:`pygame.Color`) – Scrollbar color
            - ``scrollbar_cursor``              (int, :py:class:`pygame.cursors.Cursor`, None) – Cursor of the scrollbars if the mouse is placed over
            - ``scrollbar_shadow_color``        (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the shadow of each scrollbar
            - ``scrollbar_shadow_offset``       (int) – Offset of the scrollbar shadow in px
            - ``scrollbar_shadow_position``     (str) – Position of the scrollbar shadow. See :py:mod:`pygame_menu.locals`
            - ``scrollbar_shadow``              (bool) – Indicate if a shadow is drawn on each scrollbar
            - ``scrollbar_slider_color``        (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the sliders
            - ``scrollbar_slider_hover_color``  (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the slider if hovered or clicked
            - ``scrollbar_slider_pad``          (int, float) – Space between slider and scrollbars borders in px
            - ``scrollbar_thick``               (int) – Scrollbar thickness in px
            - ``scrollbars``                    (str) – Scrollbar position. See :py:mod:`pygame_menu.locals`
            - ``shadow_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the widget shadow
            - ``shadow_radius``                 (int) - Border radius of the shadow
            - ``shadow_type``                   (str) - Shadow type, it can be ``'rectangular'`` or ``'ellipse'``
            - ``shadow_width``                  (int) - Width of the shadow. If ``0`` the shadow is disabled

        .. note::

            All theme-related optional kwargs use the default Menu theme if not
            defined.

        .. note::

            If vertical frame contains a scrollarea (setting ``max_height`` or
            ``max_width`` less than size) padding will be set at zero.

        .. note::

            Packing applies a virtual translation to the widget, previous translation
            is not modified.

        .. note::

            Widget floating is also considered within frames. If a widget is
            floating, it does not add any size to the respective positioning.

        .. note::

            The Frame size created with this method does consider the padding. Thus,
            if Frame is created with ``width=100``, ``height=200`` and ``padding=25``
            the final internal size is ``width=50`` and ``height=150``.

        .. note::

            This is applied only to the base Menu (not the currently displayed,
            stored in ``_current`` pointer); for such behaviour apply to
            :py:meth:`pygame_menu.menu.Menu.get_current` object.

        :param width: Frame width in px
        :param height: Frame height in px
        :param frame_id: ID of the vertical frame
        :param kwargs: Optional keyword arguments
        :return: Widget object
        :rtype: :py:class:`pygame_menu.widgets.Frame`
        """
        return self._frame(width, height, ORIENTATION_VERTICAL, frame_id, **kwargs)



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

SURFACE
Surface widget. This widget contains an external surface.
"""

__all__ = [
    'SurfaceWidget',
    'SurfaceWidgetManager'
]

import pygame
import pygame_menu

from abc import ABC
from pygame_menu.widgets.core.widget import Widget, WidgetTransformationNotImplemented, \
    AbstractWidgetManager

from pygame_menu._types import CallbackType, Optional, EventVectorType, Callable, \
    Any


# noinspection PyMissingOrEmptyDocstring
class SurfaceWidget(Widget):
    """
    Surface widget. Implements a widget from an external surface.

    .. note::

        SurfaceWidget only accepts translation transformation.

    :param surface: Pygame surface object
    :param surface_id: Surface ID
    :param onselect: Function when selecting the widget
    """
    _surface_obj: 'pygame.Surface'

    def __init__(
        self,
        surface: 'pygame.Surface',
        surface_id: str = '',
        onselect: CallbackType = None
    ) -> None:
        assert isinstance(surface, pygame.Surface)
        assert isinstance(surface_id, str)

        super(SurfaceWidget, self).__init__(
            onselect=onselect,
            widget_id=surface_id
        )
        self._surface_obj = surface

    def set_title(self, title: str) -> 'SurfaceWidget':
        return self

    def set_surface(self, surface: 'pygame.Surface') -> 'SurfaceWidget':
        """
        Update the widget surface.
        
        :param surface: New surface
        :return: Self reference
        """
        assert isinstance(surface, pygame.Surface)
        self._surface_obj = surface
        self._render()
        self.force_menu_surface_update()
        return self

    def _apply_font(self) -> None:
        pass

    def scale(self, *args, **kwargs) -> 'SurfaceWidget':
        raise WidgetTransformationNotImplemented()

    def resize(self, *args, **kwargs) -> 'SurfaceWidget':
        raise WidgetTransformationNotImplemented()

    def set_max_width(self, *args, **kwargs) -> 'SurfaceWidget':
        raise WidgetTransformationNotImplemented()

    def set_max_height(self, *args, **kwargs) -> 'SurfaceWidget':
        raise WidgetTransformationNotImplemented()

    def rotate(self, *args, **kwargs) -> 'SurfaceWidget':
        raise WidgetTransformationNotImplemented()

    def flip(self, *args, **kwargs) -> 'SurfaceWidget':
        raise WidgetTransformationNotImplemented()

    def _draw(self, surface: 'pygame.Surface') -> None:
        surface.blit(self._surface_obj, self._rect.topleft)

    def get_surface(self) -> 'pygame.Surface':
        return self._surface_obj

    def _render(self) -> Optional[bool]:
        self._rect.width, self._rect.height = self._surface_obj.get_size()
        return

    def update(self, events: EventVectorType) -> bool:
        self.apply_update_callbacks(events)
        for event in events:
            if self._check_mouseover(event):
                break
        return False


class SurfaceWidgetManager(AbstractWidgetManager, ABC):
    """
    SurfaceWidget manager.
    """

    def surface(
        self,
        surface: 'pygame.Surface',
        surface_id: str = '',
        onselect: Optional[Callable[[bool, 'Widget', 'pygame_menu.Menu'], Any]] = None,
        selectable: bool = False,
        **kwargs
    ) -> 'pygame_menu.widgets.SurfaceWidget':
        """
        Add a surface widget to the Menu.

        If ``onselect`` is defined, the callback is executed as follows, where
        ``selected`` is a boolean representing the selected status:

        .. code-block:: python

            onselect(selected, widget, menu)

        kwargs (Optional)
            - ``align``                         (str) – Widget `alignment <https://pygame-menu.readthedocs.io/en/latest/_source/themes.html#alignment>`_
            - ``background_color``              (tuple, list, str, int, :py:class:`pygame.Color`, :py:class:`pygame_menu.baseimage.BaseImage`) – Color of the background. ``None`` for no-color
            - ``background_inflate``            (tuple, list) – Inflate background on x-axis and y-axis (x, y) in px
            - ``border_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Widget border color. ``None`` for no-color
            - ``border_inflate``                (tuple, list) – Widget border inflate on x-axis and y-axis (x, y) in px
            - ``border_position``               (str, tuple, list) – Widget border positioning. It can be a single position, or a tuple/list of positions. Only are accepted: north, south, east, and west. See :py:mod:`pygame_menu.locals`
            - ``border_width``                  (int) – Border width in px. If ``0`` disables the border
            - ``cursor``                        (int, :py:class:`pygame.cursors.Cursor`, None) – Cursor of the widget if the mouse is placed over
            - ``float``                         (bool) - If ``True`` the widget don't contribute width/height to the Menu widget positioning computation, and don't add one unit to the rows
            - ``float_origin_position``         (bool) - If ``True`` the widget position is set to the top-left position of the Menu if the widget is floating
            - ``margin``                        (tuple, list) – Widget (left, bottom) margin in px
            - ``padding``                       (int, float, tuple, list) – Widget padding according to CSS rules. General shape: (top, right, bottom, left)
            - ``selection_color``               (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the selected widget; only affects the font color
            - ``selection_effect``              (:py:class:`pygame_menu.widgets.core.Selection`) – Widget selection effect. Applied only if ``selectable`` is ``True``
            - ``shadow_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the widget shadow
            - ``shadow_radius``                 (int) - Border radius of the shadow
            - ``shadow_type``                   (str) - Shadow type, it can be ``'rectangular'`` or ``'ellipse'``
            - ``shadow_width``                  (int) - Width of the shadow. If ``0`` the shadow is disabled

        .. note::

            All theme-related optional kwargs use the default Menu theme if not
            defined.

        .. note::

            This is applied only to the base Menu (not the currently displayed,
            stored in ``_current`` pointer); for such behaviour apply to
            :py:meth:`pygame_menu.menu.Menu.get_current` object.

        :param surface: Pygame surface object
        :param surface_id: Surface ID
        :param onselect: Callback executed when selecting the widget; only executed if ``selectable`` is ``True``
        :param selectable: Surface accepts user selection
        :param kwargs: Optional keyword arguments
        :return: Widget object
        :rtype: :py:class:`pygame_menu.widgets.SurfaceWidget`
        """
        assert isinstance(selectable, bool)

        # Remove invalid keys from kwargs
        for key in list(kwargs.keys()):
            if key not in ('align', 'background_color', 'background_inflate',
                           'border_color', 'border_inflate', 'border_width',
                           'cursor', 'margin', 'padding', 'selection_color',
                           'selection_effect', 'border_position', 'float',
                           'float_origin_position', 'shadow_color', 'shadow_radius',
                           'shadow_type', 'shadow_width'):
                kwargs.pop(key, None)

        # Filter widget attributes to avoid passing them to the callbacks
        attributes = self._filter_widget_attributes(kwargs)

        widget = SurfaceWidget(
            surface=surface,
            surface_id=surface_id,
            onselect=onselect
        )
        widget.is_selectable = selectable

        self._check_kwargs(kwargs)
        self._configure_widget(widget=widget, **attributes)
        self._append_widget(widget)

        return widget



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

EXAMPLE - CALCULATOR
Simple calculator app.
"""

__all__ = ['main']

import pygame
import pygame_menu
from pygame_menu.examples import create_example_window

from typing import Union, List


class CalculatorApp(object):
    """
    Simple calculator app.
    """
    op: str  # Operation
    prev: str  # Prev value
    curr: str  # Current value
    menu: 'pygame_menu.Menu'
    screen: 'pygame_menu.widgets.Label'
    surface: 'pygame.Surface'

    # noinspection PyArgumentEqualDefault
    def __init__(self) -> None:
        self.surface = create_example_window('Example - Calculator', (320, 480))

        # Configure theme
        theme = pygame_menu.Theme()

        theme.background_color = (43, 43, 43)
        theme.title_background_color = (43, 43, 43)
        theme.title_bar_style = pygame_menu.widgets.MENUBAR_STYLE_SIMPLE
        theme.title_close_button_cursor = pygame_menu.locals.CURSOR_HAND
        theme.title_font_size = 35
        theme.widget_alignment = pygame_menu.locals.ALIGN_LEFT
        theme.widget_background_color = None
        theme.widget_font = pygame_menu.font.FONT_DIGITAL
        theme.widget_font_color = (255, 255, 255)
        theme.widget_font_size = 40
        theme.widget_padding = 0
        theme.widget_selection_effect = \
            pygame_menu.widgets.HighlightSelection(1, 0, 0).set_color((120, 120, 120))

        self.menu = pygame_menu.Menu('', 320, 480,
                                     center_content=False,
                                     mouse_motion_selection=True,
                                     onclose=pygame_menu.events.EXIT,
                                     overflow=False,
                                     theme=theme,
                                     )
        menu_deco = self.menu.get_scrollarea().get_decorator()

        # Add the layout
        self.menu.add.vertical_margin(40)
        menu_deco.add_rectangle(10, 88, 300, 55, (60, 63, 65), use_center_positioning=False)
        self.screen = self.menu.add.label('0', background_color=None, margin=(10, 0),
                                          selectable=True, selection_effect=None)
        self.menu.add.vertical_margin(20)

        cursor = pygame_menu.locals.CURSOR_HAND

        # Add horizontal frames
        f1 = self.menu.add.frame_h(299, 54, margin=(10, 0))
        b1 = f1.pack(self.menu.add.button('1', lambda: self._press(1), cursor=cursor))
        b2 = f1.pack(self.menu.add.button('2', lambda: self._press(2), cursor=cursor), align=pygame_menu.locals.ALIGN_CENTER)
        b3 = f1.pack(self.menu.add.button('3', lambda: self._press(3), cursor=cursor), align=pygame_menu.locals.ALIGN_RIGHT)
        self.menu.add.vertical_margin(10)

        f2 = self.menu.add.frame_h(299, 54, margin=(10, 0))
        b4 = f2.pack(self.menu.add.button('4', lambda: self._press(4), cursor=cursor))
        b5 = f2.pack(self.menu.add.button('5', lambda: self._press(5), cursor=cursor), align=pygame_menu.locals.ALIGN_CENTER)
        b6 = f2.pack(self.menu.add.button('6', lambda: self._press(6), cursor=cursor), align=pygame_menu.locals.ALIGN_RIGHT)
        self.menu.add.vertical_margin(10)

        f3 = self.menu.add.frame_h(299, 54, margin=(10, 0))
        b7 = f3.pack(self.menu.add.button('7', lambda: self._press(7), cursor=cursor))
        b8 = f3.pack(self.menu.add.button('8', lambda: self._press(8), cursor=cursor), align=pygame_menu.locals.ALIGN_CENTER)
        b9 = f3.pack(self.menu.add.button('9', lambda: self._press(9), cursor=cursor), align=pygame_menu.locals.ALIGN_RIGHT)
        self.menu.add.vertical_margin(10)

        f4 = self.menu.add.frame_h(299, 54, margin=(10, 0))
        b0 = f4.pack(self.menu.add.button('0', lambda: self._press(0), cursor=cursor))
        b_plus = f4.pack(self.menu.add.button('+', lambda: self._press('+'), cursor=cursor), align=pygame_menu.locals.ALIGN_CENTER)
        b_minus = f4.pack(self.menu.add.button('-', lambda: self._press('-'), cursor=cursor), align=pygame_menu.locals.ALIGN_RIGHT)
        self.menu.add.vertical_margin(10)

        f5 = self.menu.add.frame_h(299, 54, margin=(10, 0))
        b_times = f5.pack(self.menu.add.button('x', lambda: self._press('x'), cursor=cursor))
        b_div = f5.pack(self.menu.add.button('/', lambda: self._press('/'), cursor=cursor), align=pygame_menu.locals.ALIGN_CENTER)
        beq = f5.pack(self.menu.add.button('=', lambda: self._press('='), cursor=cursor), align=pygame_menu.locals.ALIGN_RIGHT)

        # Add decorator for each object
        for widget in (b1, b2, b3, b4, b5, b6, b7, b8, b9, b0, beq, b_plus, b_minus, b_times, b_div):
            w_deco = widget.get_decorator()
            if widget != beq:
                w_deco.add_rectangle(-37, -27, 74, 54, (15, 15, 15))
                on_layer = w_deco.add_rectangle(-37, -27, 74, 54, (84, 84, 84))
            else:
                w_deco.add_rectangle(-37, -27, 74, 54, (38, 96, 103))
                on_layer = w_deco.add_rectangle(-37, -27, 74, 54, (40, 171, 187))
            w_deco.disable(on_layer)
            widget.set_attribute('on_layer', on_layer)

            def widget_select(sel: bool, wid: 'pygame_menu.widgets.Widget', _):
                """
                Function triggered if widget is selected
                """
                lay = wid.get_attribute('on_layer')
                if sel:
                    wid.get_decorator().enable(lay)
                else:
                    wid.get_decorator().disable(lay)

            widget.set_onselect(widget_select)
            widget.set_padding((2, 19, 0, 23))
            widget._keyboard_enabled = False

        self.prev = ''
        self.curr = ''
        self.op = ''

        self.menu.set_onupdate(self.process_events)
        self.menu.set_onwindowmouseleave(lambda m: self.screen.select(update_menu=True))

    def process_events(self, events: List['pygame.event.Event'], _=None) -> None:
        """
        Process events from user.
        """
        for event in events:
            if event.type == pygame.KEYDOWN:
                # noinspection PyUnresolvedReferences
                if event.key == pygame.K_0:
                    self._press(0)
                elif event.key == pygame.K_1:
                    self._press(1)
                elif event.key == pygame.K_2:
                    self._press(2)
                elif event.key == pygame.K_3:
                    self._press(3)
                elif event.key == pygame.K_4:
                    self._press(4)
                elif event.key == pygame.K_5:
                    self._press(5)
                elif event.key == pygame.K_6:
                    self._press(6)
                elif event.key == pygame.K_7:
                    self._press(7)
                elif event.key == pygame.K_8:
                    self._press(8)
                elif event.key == pygame.K_9:
                    self._press(9)
                elif event.key == pygame.K_PLUS:
                    self._press('+')
                elif event.key == pygame.K_MINUS:
                    self._press('-')
                elif event.key == pygame.K_SLASH or (hasattr(pygame, 'K_PERCENT') and event.key == pygame.K_PERCENT):
                    self._press('/')
                elif event.key == pygame.K_ASTERISK or event.key == pygame.K_x:
                    self._press('x')
                elif event.key == pygame.K_EQUALS or event.key == pygame.K_RETURN:
                    self._press('=')
                elif event.key == pygame.K_BACKSPACE:
                    self._press('=')
                    self._press('=')

    def _operate(self) -> Union[int, float]:
        """
        Operate current and previous values.

        :return: Operation result
        """
        a = 0 if self.curr == '' else float(self.curr)
        b = 0 if self.prev == '' else float(self.prev)
        c = 0
        if self.op == '+':
            c = a + b
        elif self.op == '-':
            c = b - a
        elif self.op == 'x':
            c = a * b
        elif self.op == '/':
            if a != 0:
                c = b / a
            else:
                self.screen.set_title('Error')
        return int(c)

    def _press(self, digit: Union[int, str]) -> None:
        """
        Press calculator digit.

        :param digit: Number or symbol
        """
        if digit in ('+', '-', 'x', '/'):
            if self.curr != '':
                if self.op != '':
                    self.prev = str(self._operate())
                else:
                    self.prev = self.curr
                self.curr = ''
            self.op = digit
            if len(self.prev) <= 8:
                self.screen.set_title(self.prev + self.op)
            else:
                self.screen.set_title('Ans' + self.op)
        elif digit == '=':
            if self.prev == '':
                self.curr = ''
                self.screen.set_title('0')
                return
            c = self._operate()
            self.screen.set_title(str(c))
            if len(str(c)) > 8:
                c = 0
                self.screen.set_title('Overflow')
            self.prev = ''
            self.curr = str(c)
            self.op = ''
        else:
            if self.op == '':
                if len(self.prev) <= 7:
                    self.prev += str(digit)
                    self.prev = self._format(self.prev)
                self.screen.set_title(self.prev)
            else:
                if len(self.curr) <= 7:
                    self.curr += str(digit)
                    self.curr = self._format(self.curr)
                self.screen.set_title(self.curr)

    @staticmethod
    def _format(x: str) -> str:
        """
        Format number.

        :param x: Number
        :return: Str
        """
        try:
            if int(x) == float(x):
                return str(int(x))
        except ValueError:
            pass
        try:
            x = float(x)
        except ValueError:
            pass

        return str(round(int(x), 0))

    def mainloop(self, test: bool) -> None:
        """
        App mainloop.

        :param test: Test status
        """
        self.menu.mainloop(self.surface, disable_loop=test)


def main(test: bool = False) -> 'CalculatorApp':
    """
    Main function.

    :param test: Indicate function is being tested
    :return: App object
    """
    app = CalculatorApp()
    app.mainloop(test)
    return app


if __name__ == '__main__':
    main()



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

EXAMPLE - USE SCROLLBAR WIDGET
Shows how the ScrollBar can be used on a surface.
"""

__all__ = ['main']

import pygame
import pygame_menu

from pygame_menu.examples import create_example_window
from pygame_menu.utils import make_surface
from pygame_menu.widgets import ScrollBar


def make_world(width: int, height: int) -> 'pygame.Surface':
    """
    Create a test surface.

    :param width: Width in pixels
    :param height: Height in pixels
    :return: World surface
    """
    world = make_surface(width, height)
    world.fill((200, 200, 200))

    color = [70, 20, 20]
    max_x = len(list(range(100, width, 200)))
    max_y = len(list(range(100, height, 200)))
    number_x = 0
    for x in range(100, width, 200):
        number_y = 0
        for y in range(100, height, 200):
            if number_x in (0, max_x - 1) or number_y in (0, max_y - 1):
                # White circles to delimit world boundaries
                # noinspection PyArgumentList
                pygame.draw.circle(world, (255, 255, 255), (x, y), 100, 10)
            else:
                # noinspection PyArgumentList
                pygame.draw.circle(world, color, (x, y), 100, 10)
                if color[0] + 15 < 255:
                    color[0] += 15
                elif color[1] + 15 < 255:
                    color[1] += 15
                else:
                    color[2] += 15
            number_y += 1
        number_x += 1

    return world


def h_changed(value: int) -> None:
    """
    :param value: Value data
    """
    print('Horizontal position changed:', value)


def v_changed(value: int) -> None:
    """
    :param value: Value data
    """
    print('Vertical position changed:', value)


def main(test: bool = False) -> None:
    """
    Main function.

    :param test: Indicate function is being tested
    """
    scr_size = (480, 480)
    screen = create_example_window('Example - Scrollbar', scr_size)
    world = make_world(int(scr_size[0] * 4), scr_size[1] * 3)
    screen.fill((120, 90, 130))

    thick_h = 20
    thick_v = 40

    # Horizontal ScrollBar
    sb_h = ScrollBar(
        length=scr_size[0] - thick_v,
        values_range=(50, world.get_width() - scr_size[0] + thick_v),
        slider_pad=2,
        page_ctrl_thick=thick_h,
        onchange=h_changed
    )
    sb_h.set_shadow(
        color=(0, 0, 0),
        position=pygame_menu.locals.POSITION_SOUTHEAST
    )
    sb_h.set_controls(False)
    sb_h.set_position(0, scr_size[1] - thick_h)
    sb_h.set_page_step(scr_size[0] - thick_v)

    # Vertical ScrollBar
    sb_v = ScrollBar(
        length=scr_size[1] - thick_h,
        values_range=(0, world.get_height() - scr_size[1] + thick_h),
        orientation=pygame_menu.locals.ORIENTATION_VERTICAL,
        slider_pad=6,
        slider_color=(135, 193, 180),
        slider_hover_color=(180, 180, 180),
        page_ctrl_thick=thick_v,
        page_ctrl_color=(253, 246, 220),
        onchange=v_changed
    )
    sb_v.set_shadow(
        color=(52, 54, 56),
        position=pygame_menu.locals.POSITION_NORTHWEST,
        offset=4
    )
    sb_v.set_controls(False)
    sb_v.set_position(scr_size[0] - thick_v, 0)
    sb_v.set_page_step(scr_size[1] - thick_h)
    clock = pygame.time.Clock()

    # -------------------------------------------------------------------------
    # Main loop
    # -------------------------------------------------------------------------
    while True:

        # Clock tick
        clock.tick(60)

        # Application events
        events = pygame.event.get()
        for event in events:
            if event.type == pygame.QUIT:
                exit()

            if event.type == pygame.KEYDOWN and event.key == pygame.K_h:
                sb_h.set_value(100)

            if event.type == pygame.KEYDOWN and event.key == pygame.K_v:
                sb_v.set_value(200)

            sb_h.update([event])
            sb_h.draw(screen)
            sb_v.update([event])
            sb_v.draw(screen)

        trunc_world_orig = (sb_h.get_value(), sb_v.get_value())
        trunc_world = (scr_size[0] - thick_v, scr_size[1] - thick_h)

        # noinspection PyTypeChecker
        screen.blit(world, (0, 0), (trunc_world_orig, trunc_world))
        pygame.display.update()

        # At first loop returns
        if test:
            break


if __name__ == '__main__':
    main()



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

MENUBAR
MenuBar class to display the Menu title.
"""
# File constants no. 1000

__all__ = [

    # Main class
    'MenuBar',

    # Menubar styles
    'MENUBAR_STYLE_ADAPTIVE',
    'MENUBAR_STYLE_SIMPLE',
    'MENUBAR_STYLE_TITLE_ONLY',
    'MENUBAR_STYLE_TITLE_ONLY_DIAGONAL',
    'MENUBAR_STYLE_NONE',
    'MENUBAR_STYLE_UNDERLINE',
    'MENUBAR_STYLE_UNDERLINE_TITLE',

    # Custom types
    'MenuBarStyleModeType'

]

import pygame
import pygame.gfxdraw as gfxdraw

from pygame_menu.locals import FINGERUP, POSITION_EAST, POSITION_WEST, POSITION_NORTH, \
    POSITION_SOUTH
from pygame_menu.utils import assert_color, get_finger_pos, warn
from pygame_menu.widgets.core.widget import Widget, WidgetTransformationNotImplemented

from pygame_menu._types import Tuple, CallbackType, Tuple2IntType, Any, \
    Optional, NumberInstance, ColorInputType, EventVectorType, VectorInstance, \
    List, ColorType, NumberType

# Menubar styles
MENUBAR_STYLE_ADAPTIVE = 1000
MENUBAR_STYLE_SIMPLE = 1001
MENUBAR_STYLE_TITLE_ONLY = 1002
MENUBAR_STYLE_TITLE_ONLY_DIAGONAL = 1003
MENUBAR_STYLE_NONE = 1004
MENUBAR_STYLE_UNDERLINE = 1005
MENUBAR_STYLE_UNDERLINE_TITLE = 1006

# Menubar operation modes
_MODE_CLOSE = 1020
_MODE_BACK = 1021

# Custom types
MenuBarStyleModeType = int


# noinspection PyMissingOrEmptyDocstring
class MenuBar(Widget):
    """
    MenuBar widget.

    .. note::

        MenuBar only accepts translation transformation.

    :param title: Title of the menubar
    :param width: Width of the widget, generally the same as the width of the menu
    :param background_color: Background color
    :param menubar_id: ID of the MenuBar
    :param back_box: Draw a back-box button on header
    :param back_box_background_color: Back-box button color
    :param mode: Mode of drawing the bar
    :param modify_scrollarea: If ``True`` it modifies the scrollbars of the scrollarea depending on the bar mode
    :param offsetx: Offset x-position of title in px
    :param offsety: Offset y-position of title in px
    :param onreturn: Callback when pressing the back-box button
    :param args: Optional arguments for callbacks
    :param kwargs: Optional keyword arguments for callbacks
    """
    _backbox: bool
    _backbox_background_color: ColorType
    _backbox_border_width: int
    _backbox_pos: Any
    _backbox_rect: Optional['pygame.Rect']
    _box_mode: int
    _modify_scrollarea: bool
    _offsetx: NumberType
    _offsety: NumberType
    _polygon_pos: Any
    _scrollbar_deltas: List[Tuple[int, Tuple2IntType]]
    _style: int
    _width: int
    fixed: bool

    def __init__(
        self,
        title: Any,
        width: NumberType,
        background_color: ColorInputType,
        menubar_id: str = '',
        back_box: bool = False,
        back_box_background_color: ColorInputType = (0, 0, 0),
        mode: MenuBarStyleModeType = MENUBAR_STYLE_ADAPTIVE,
        modify_scrollarea: bool = True,
        offsetx: NumberType = 0,
        offsety: NumberType = 0,
        onreturn: CallbackType = None,
        *args,
        **kwargs
    ) -> None:
        assert isinstance(width, NumberInstance)
        assert isinstance(back_box, bool)

        assert width > 0, 'width must be greater or equal than zero'

        background_color = assert_color(background_color)
        back_box_background_color = assert_color(back_box_background_color)

        # MenuBar has no ID
        super(MenuBar, self).__init__(
            args=args,
            kwargs=kwargs,
            onreturn=onreturn,
            title=title,
            widget_id=menubar_id
        )

        self._backbox = back_box
        self._backbox_background_color = back_box_background_color
        self._backbox_border_width = 1  # px
        self._backbox_pos = None
        self._backbox_rect = None
        self._background_color = background_color
        self._box_mode = 0
        self._modify_scrollarea = modify_scrollarea
        self._mouseover_check_rect = lambda: self._backbox_rect
        self._offsetx = 0
        self._offsety = 0
        self._polygon_pos = None
        # north, east, south, west
        self._scrollbar_deltas = [(0, (0, 0)), (0, (0, 0)), (0, (0, 0)), (0, (0, 0))]
        self._style = mode
        self._title = ''
        self._width = int(width)

        self.set_title(title, offsetx, offsety)

        # Public's
        self.is_selectable = False
        self.fixed = True

    def _apply_font(self) -> None:
        pass

    def set_padding(self, *args, **kwargs) -> 'MenuBar':
        return self

    def scale(self, *args, **kwargs) -> 'MenuBar':
        raise WidgetTransformationNotImplemented()

    def resize(self, *args, **kwargs) -> 'MenuBar':
        raise WidgetTransformationNotImplemented()

    def set_max_width(self, *args, **kwargs) -> 'MenuBar':
        raise WidgetTransformationNotImplemented()

    def set_max_height(self, *args, **kwargs) -> 'MenuBar':
        raise WidgetTransformationNotImplemented()

    def rotate(self, *args, **kwargs) -> 'MenuBar':
        raise WidgetTransformationNotImplemented()

    def flip(self, *args, **kwargs) -> 'MenuBar':
        raise WidgetTransformationNotImplemented()

    def set_selection_effect(self, *args, **kwargs) -> 'MenuBar':
        return self

    def set_border(self, *args, **kwargs) -> 'MenuBar':
        return self

    def _check_title_color(self, background_menu: bool) -> None:
        """
        Performs title color and prints a warning if the color is similar to the
        background.
        """
        if background_menu and self._menu is not None:
            c_back = self._menu.get_theme().background_color
        else:
            c_back = self._background_color
        if not isinstance(c_back, VectorInstance):  # If is color
            return
        tol = 5
        c_dif_1 = abs(c_back[0] - self._font_color[0])
        c_dif_2 = abs(c_back[1] - self._font_color[1])
        c_dif_3 = abs(c_back[2] - self._font_color[2])
        if self._verbose and c_dif_1 < tol and c_dif_2 < tol and c_dif_3 < tol:
            warn(
                'title font color {0} is {3} to the {1} background color {2}, '
                'consider editing your Theme'.format(
                    self._font_color,
                    'menu' if background_menu else 'title',
                    c_back,
                    'equal' if c_dif_1 == c_dif_2 == c_dif_3 == 0 else 'similar'
                )
            )

    def get_title_offset(self) -> Tuple2IntType:
        """
        Return the title offset on x-axis and y-axis (x, y) in px.

        :return: Title offset
        """
        return int(self._offsetx), int(self._offsety)

    def set_backbox_border_width(self, width: int) -> None:
        """
        Set backbox border width in px.

        :param width: Width in px
        """
        assert isinstance(width, int)
        assert width > 0
        self._backbox_border_width = width

    def _draw_background_color(self, *args, **kwargs) -> None:
        pass

    def _draw_border(self, *args, **kwargs) -> None:
        pass

    def _backbox_visible(self) -> bool:
        """
        Return ``True`` if backbox is visible.

        :return: Bool
        """
        # The following check belongs to the case if the Menu displays a "x" button
        # to close the Menu, but onclose Menu method is None (Nothing is executed),
        # then the button will not be displayed
        # noinspection PyProtectedMember
        return (self._mouse_enabled or self._touchscreen_enabled) and self._backbox and \
            not (self._box_mode == _MODE_CLOSE and
                 self._menu is not None and
                 self._menu._onclose is None)

    def _draw(self, surface: 'pygame.Surface') -> None:
        if len(self._polygon_pos) > 2:
            gfxdraw.filled_polygon(surface, self._polygon_pos, self._background_color)

        # Draw backbox if enabled
        if self._backbox_visible():
            # noinspection PyArgumentList
            pygame.draw.rect(surface, self._backbox_background_color, self._backbox_rect, self._backbox_border_width)
            pygame.draw.polygon(surface, self._backbox_background_color, self._backbox_pos)

        surface.blit(self._surface,
                     (self._rect.topleft[0] + self._offsetx,
                      self._rect.topleft[1] + self._offsety))

    def get_scrollbar_style_change(self, position: str) -> Tuple[int, Tuple2IntType]:
        """
        Return scrollbar change (width, position) depending on the style of the
        menubar.

        :param position: Position of the scrollbar
        :return: Change in length and position in px
        """
        self._render()
        if not self._modify_scrollarea or not self.is_visible():
            return 0, (0, 0)
        if not self.fixed or self.is_floating():
            if self._style == MENUBAR_STYLE_ADAPTIVE:
                if position == POSITION_EAST:
                    t = self._polygon_pos[4][1] - self._polygon_pos[2][1]
                    return t, (0, -t)
            return 0, (0, 0)
        if position == POSITION_NORTH:
            return self._scrollbar_deltas[0]
        elif position == POSITION_EAST:
            return self._scrollbar_deltas[1]
        elif position == POSITION_SOUTH:
            return self._scrollbar_deltas[2]
        elif position == POSITION_WEST:
            return self._scrollbar_deltas[3]
        return 0, (0, 0)

    def _render(self) -> Optional[bool]:
        if self._menu is None:
            return

        # noinspection PyProtectedMember
        menu_prev_condition = not self._menu or not self._menu._top or not self._menu._top._prev

        if not self._render_hash_changed(self._menu.get_id(), self._rect.x, self._rect.y, self._title, self._width,
                                         self._visible, self._font_selected_color, menu_prev_condition):
            return True

        # Update box mode
        if menu_prev_condition:
            self._box_mode = _MODE_CLOSE
        else:
            self._box_mode = _MODE_BACK

        self._surface = self._render_string(self._title, self._font_selected_color)
        self._rect.width, self._rect.height = self._surface.get_size()
        self._apply_transforms()  # Rotation does not affect rect size

        dy = 0

        if self._style == MENUBAR_STYLE_ADAPTIVE:
            """
            A-------------------B                  D-E: 25 dx
            |****             x | *0,6 height
            |      D------------C
            F----E/
            """
            a = self._rect.x, self._rect.y
            b = self._rect.x + self._width - 1, self._rect.y
            c = self._rect.x + self._width - 1, self._rect.y + self._rect.height * 0.6
            d = self._rect.x + self._rect.width + 25 + self._offsetx, self._rect.y + self._rect.height * 0.6
            e = self._rect.x + self._rect.width + 5 + self._offsetx, self._rect.y + self._rect.height
            f = self._rect.x, self._rect.y + self._rect.height
            self._polygon_pos = a, b, c, d, e, f
            cross_size = int(self._rect.height * 0.6)
            self._scrollbar_deltas = [(0, (0, self._rect.height)),
                                      (-cross_size, (0, cross_size)),
                                      (0, (0, 0)),
                                      (-self._rect.height, (0, self._rect.height))]
            self._check_title_color(background_menu=False)

        elif self._style == MENUBAR_STYLE_SIMPLE:
            """
            A-------------------B
            |****             x | *1,0 height
            D-------------------C
            """
            a = self._rect.x, self._rect.y
            b = self._rect.x + self._width - 1, self._rect.y
            c = self._rect.x + self._width - 1, self._rect.y + self._rect.height
            d = self._rect.x, self._rect.y + self._rect.height
            self._polygon_pos = a, b, c, d
            cross_size = int(self._rect.height * self._backbox_visible())
            self._scrollbar_deltas = [(0, (0, self._rect.height)),
                                      (-self._rect.height, (0, self._rect.height)),
                                      (0, (0, 0)),
                                      (-self._rect.height, (0, self._rect.height))]
            self._check_title_color(background_menu=False)

        elif self._style == MENUBAR_STYLE_TITLE_ONLY:
            """
            A-----B
            | *** |           x        *0,6 height
            D-----C
            """
            a = self._rect.x, self._rect.y
            b = self._rect.x + self._rect.width + 5 + self._offsetx, self._rect.y
            c = self._rect.x + self._rect.width + 5 + self._offsetx, self._rect.y + self._rect.height
            d = self._rect.x, self._rect.y + self._rect.height
            self._polygon_pos = a, b, c, d
            cross_size = int(self._rect.height * 0.6 * self._backbox_visible())
            self._scrollbar_deltas = [(0, (0, self._rect.height)),
                                      (-cross_size, (0, cross_size)),
                                      (0, (0, 0)),
                                      (-self._rect.height, (0, self._rect.height))]
            self._check_title_color(background_menu=False)

        elif self._style == MENUBAR_STYLE_TITLE_ONLY_DIAGONAL:
            """
            A--------B
            | **** /          x        *0,6 height
            D-----C
            """
            a = self._rect.x, self._rect.y
            b = self._rect.x + self._rect.width + 25 + self._offsetx, self._rect.y
            c = self._rect.x + self._rect.width + 5 + self._offsetx, self._rect.y + self._rect.height
            d = self._rect.x, self._rect.y + self._rect.height
            self._polygon_pos = a, b, c, d
            cross_size = int(self._rect.height * 0.6 * self._backbox_visible())
            self._scrollbar_deltas = [(0, (0, self._rect.height)),
                                      (-cross_size, (0, cross_size)),
                                      (0, (0, 0)),
                                      (-self._rect.height, (0, self._rect.height))]
            self._check_title_color(background_menu=False)

        elif self._style == MENUBAR_STYLE_NONE:
            """
            A------------------B
             ****             x        *0,6 height
            """
            a = self._rect.x, self._rect.y
            b = self._rect.x + self._width - 1, self._rect.y
            self._polygon_pos = a, b
            cross_size = int(self._rect.height * 0.6 * self._backbox_visible())
            self._scrollbar_deltas = [(0, (0, self._rect.height)),
                                      (-cross_size, (0, cross_size)),
                                      (0, (0, 0)),
                                      (-self._rect.height, (0, self._rect.height))]
            self._check_title_color(background_menu=True)

        elif self._style == MENUBAR_STYLE_UNDERLINE:
            """
             ****             x
            A-------------------B      *0,09 height
            D-------------------C
            """
            # dy = 0
            a = self._rect.x, self._rect.y + 0.91 * self._rect.height + dy
            b = self._rect.x + self._width - 1, self._rect.y + 0.91 * self._rect.height + dy
            c = self._rect.x + self._width - 1, self._rect.y + self._rect.height + dy
            d = self._rect.x, self._rect.y + self._rect.height + dy
            self._polygon_pos = a, b, c, d
            cross_size = int(0.6 * self._rect.height * self._backbox_visible())
            self._scrollbar_deltas = [(0, (0, self._rect.height)),
                                      (-self._rect.height, (0, self._rect.height)),
                                      (0, (0, 0)),
                                      (-self._rect.height, (0, self._rect.height))]
            self._check_title_color(background_menu=True)

        elif self._style == MENUBAR_STYLE_UNDERLINE_TITLE:
            """
             ****               x
            A----B                     *0,09 height
            D----C
            """
            # dy = 3
            a = self._rect.x, self._rect.y + 0.91 * self._rect.height + dy
            b = self._rect.x + self._rect.width + 5 + self._offsetx, self._rect.y + 0.91 * self._rect.height + dy
            c = self._rect.x + self._rect.width + 5 + self._offsetx, self._rect.y + self._rect.height + dy
            d = self._rect.x, self._rect.y + self._rect.height + dy
            self._polygon_pos = a, b, c, d
            cross_size = int(0.6 * self._rect.height * self._backbox_visible())
            self._scrollbar_deltas = [(0, (0, self._rect.height)),
                                      (-cross_size, (0, cross_size)),
                                      (0, (0, 0)),
                                      (-self._rect.height, (0, self._rect.height))]
            self._check_title_color(background_menu=True)

        else:
            raise ValueError(f'invalid menubar mode {self._style}')
        self._rect.height += dy

        # Create the back box
        if self._backbox:
            backbox_margin = 4

            # Subtract the scrollarea thickness if float and enabled
            scroll_delta = 0
            if self._floating and self._menu is not None:
                scroll_delta = self._menu.get_width() - self._menu.get_width(inner=True)

            self._backbox_rect = pygame.Rect(
                int(self._rect.x + self._width - cross_size + backbox_margin - scroll_delta),
                int(self._rect.y + backbox_margin),
                int(cross_size - 2 * backbox_margin),
                int(cross_size - 2 * backbox_margin)
            )

            if self._box_mode == _MODE_CLOSE:
                # Make a cross for top Menu
                self._backbox_pos = (
                    (self._backbox_rect.left + 4, self._backbox_rect.top + 4),
                    (self._backbox_rect.centerx, self._backbox_rect.centery),
                    (self._backbox_rect.right - 4, self._backbox_rect.top + 4),
                    (self._backbox_rect.centerx, self._backbox_rect.centery),
                    (self._backbox_rect.right - 4, self._backbox_rect.bottom - 4),
                    (self._backbox_rect.centerx, self._backbox_rect.centery),
                    (self._backbox_rect.left + 4, self._backbox_rect.bottom - 4),
                    (self._backbox_rect.centerx, self._backbox_rect.centery),
                    (self._backbox_rect.left + 4, self._backbox_rect.top + 4)
                )

            elif self._box_mode == _MODE_BACK:
                # Make a back arrow for sub-menus
                self._backbox_pos = (
                    (self._backbox_rect.left + 5, self._backbox_rect.centery),
                    (self._backbox_rect.centerx, self._backbox_rect.top + 5),
                    (self._backbox_rect.centerx, self._backbox_rect.centery - 2),
                    (self._backbox_rect.right - 5, self._backbox_rect.centery - 2),
                    (self._backbox_rect.right - 5, self._backbox_rect.centery + 2),
                    (self._backbox_rect.centerx, self._backbox_rect.centery + 2),
                    (self._backbox_rect.centerx, self._backbox_rect.bottom - 5),
                    (self._backbox_rect.left + 5, self._backbox_rect.centery)
                )

    def set_title(self, title: Any, offsetx: NumberType = 0, offsety: NumberType = 0) -> 'MenuBar':
        """
        Set the menubar title.

        :param title: Menu title
        :param offsetx: Offset x-position of title in px
        :param offsety: Offset y-position of title in px
        :return: Self reference
        """
        assert isinstance(offsetx, NumberInstance)
        assert isinstance(offsety, NumberInstance)
        self._title = str(title)
        self._offsety = offsety
        self._offsetx = offsetx
        if self._menu is not None:
            self._render()
        return self

    def get_height(self, apply_padding: bool = True, apply_selection: bool = False) -> int:
        if self._floating or not self.is_visible():
            return 0
        return super(MenuBar, self).get_height(apply_padding, apply_selection)

    def update(self, events: EventVectorType) -> bool:
        self.apply_update_callbacks(events)

        if self.readonly or not self.is_visible():
            self._readonly_check_mouseover(events)
            return False

        for event in events:

            # Check mouse over
            if self._backbox_visible():
                self._check_mouseover(event)

            # User clicks/touches the backbox rect; don't consider the mouse wheel (button 4 & 5)
            if (
                event.type == pygame.MOUSEBUTTONUP and self._mouse_enabled and event.button in (1, 2, 3) or
                event.type == FINGERUP and self._touchscreen_enabled and self._menu is not None
            ):
                event_pos = get_finger_pos(self._menu, event)
                if self._backbox_visible() and self._backbox_rect.collidepoint(*event_pos):
                    if event.type == pygame.MOUSEBUTTONUP:
                        self._sound.play_click_mouse()
                    else:
                        self._sound.play_click_touch()
                    self.apply()
                    return True

            # User applies joy back button
            elif event.type == pygame.JOYBUTTONDOWN and self._joystick_enabled:
                if self._ctrl.joy_back(event, self):
                    self._sound.play_key_del()
                    self.apply()
                    return True

        return False



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

NONE
No selection effect.
"""

__all__ = ['NoneSelection']

import pygame
import pygame_menu

from pygame_menu.widgets.selection.simple import SimpleSelection


class NoneSelection(SimpleSelection):
    """
    No selection effect.
    """

    def __init__(self) -> None:
        super(NoneSelection, self).__init__()
        self.widget_apply_font_color = False

    # noinspection PyMissingOrEmptyDocstring
    def draw(self, surface: 'pygame.Surface', widget: 'pygame_menu.widgets.Widget') -> 'NoneSelection':
        return self



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

EXAMPLE - IMAGE BACKGROUND
Menu using background image + BaseImage object.
"""

__all__ = ['main']

import pygame
import pygame_menu
from pygame_menu.examples import create_example_window

from typing import Optional

# Constants and global variables
FPS = 60
WINDOW_SIZE = (640, 480)

sound: Optional['pygame_menu.sound.Sound'] = None
surface: Optional['pygame.Surface'] = None
main_menu: Optional['pygame_menu.Menu'] = None

# Load image
background_image = pygame_menu.BaseImage(
    image_path=pygame_menu.baseimage.IMAGE_EXAMPLE_WALLPAPER
)


def main_background() -> None:
    """
    Background color of the main menu, on this function user can plot
    images, play sounds, etc.
    """
    background_image.draw(surface)


def main(test: bool = False) -> None:
    """
    Main program.

    :param test: Indicate function is being tested
    """
    global main_menu
    global sound
    global surface

    # Create window
    surface = create_example_window('Example - Image Background', WINDOW_SIZE)
    clock = pygame.time.Clock()

    # Create menus: Main menu
    main_menu_theme = pygame_menu.themes.THEME_ORANGE.copy()
    main_menu_theme.set_background_color_opacity(0.5)  # 50% opacity

    main_menu = pygame_menu.Menu(
        height=WINDOW_SIZE[1] * 0.7,
        onclose=pygame_menu.events.EXIT,  # User press ESC button
        theme=main_menu_theme,
        title='Epic Menu',
        width=WINDOW_SIZE[0] * 0.8
    )

    theme_bg_image = main_menu_theme.copy()
    theme_bg_image.background_color = pygame_menu.BaseImage(
        image_path=pygame_menu.baseimage.IMAGE_EXAMPLE_CARBON_FIBER
    )
    theme_bg_image.title_font_size = 25
    menu_with_bg_image = pygame_menu.Menu(
        height=WINDOW_SIZE[1] * 0.7,
        onclose=pygame_menu.events.EXIT,
        theme=theme_bg_image,
        title='Menu with background image',
        width=WINDOW_SIZE[0] * 0.8
    )
    menu_with_bg_image.add.button('Back', pygame_menu.events.BACK)

    widget_colors_theme = pygame_menu.themes.THEME_ORANGE.copy()
    widget_colors_theme.widget_margin = (0, 10)
    widget_colors_theme.widget_padding = 0
    widget_colors_theme.widget_selection_effect.margin_xy(10, 5)
    widget_colors_theme.widget_font_size = 20
    widget_colors_theme.set_background_color_opacity(0.5)  # 50% opacity

    widget_colors = pygame_menu.Menu(
        height=WINDOW_SIZE[1] * 0.7,
        theme=widget_colors_theme,
        title='Widget backgrounds',
        width=WINDOW_SIZE[0] * 0.8
    )

    button_image = pygame_menu.BaseImage(pygame_menu.baseimage.IMAGE_EXAMPLE_CARBON_FIBER)

    widget_colors.add.button('Opaque color button',
                             background_color=(100, 100, 100))
    widget_colors.add.button('Transparent color button',
                             background_color=(50, 50, 50, 200), font_size=40)
    widget_colors.add.button('Transparent background inflate to selection effect',
                             background_color=(50, 50, 50, 200),
                             margin=(0, 15)).background_inflate_to_selection_effect()
    widget_colors.add.button('Background inflate + font background color',
                             background_color=(50, 50, 50, 200),
                             font_background_color=(200, 200, 200)
                             ).background_inflate_to_selection_effect()
    widget_colors.add.button('This inflates background to match selection effect',
                             background_color=button_image,
                             font_color=(255, 255, 255), font_size=15
                             ).selection_expand_background = True
    widget_colors.add.button('This is already inflated to match selection effect',
                             background_color=button_image,
                             font_color=(255, 255, 255), font_size=15
                             ).background_inflate_to_selection_effect()

    main_menu.add.button('Menu with background image', menu_with_bg_image)
    main_menu.add.button('Test different widget colors', widget_colors)
    main_menu.add.button('Another fancy button', lambda: print('This button has been pressed'))
    main_menu.add.button('Quit', pygame_menu.events.EXIT)

    # Main loop
    while True:

        # Tick
        clock.tick(FPS)

        # Main menu
        main_menu.mainloop(surface, main_background, disable_loop=test, fps_limit=FPS)

        # Flip surface
        pygame.display.flip()

        # At first loop returns
        if test:
            break


if __name__ == '__main__':
    main()



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

CONTROLS
Default controls of Menu object and key definition.
"""

__all__ = [

    # Joy pad
    'JOY_AXIS_X',
    'JOY_AXIS_Y',
    'JOY_BUTTON_BACK',
    'JOY_BUTTON_SELECT',
    'JOY_DEADZONE',
    'JOY_DELAY',
    'JOY_DOWN',
    'JOY_LEFT',
    'JOY_REPEAT',
    'JOY_RIGHT',
    'JOY_UP',

    # Keyboard events
    'KEY_APPLY',
    'KEY_BACK',
    'KEY_CLOSE_MENU',
    'KEY_LEFT',
    'KEY_MOVE_DOWN',
    'KEY_MOVE_UP',
    'KEY_RIGHT',

    # Controller object
    'Controller'

]

# Imports
# noinspection PyUnresolvedReferences
import pygame_menu
import pygame.locals as _locals
from pygame.event import Event as EventType
from typing import Union

WidgetType = Union['pygame_menu.Menu', 'pygame_menu.widgets.Widget']

# Joy pad
JOY_AXIS_X = 0
JOY_AXIS_Y = 1
JOY_BUTTON_BACK = 1
JOY_BUTTON_SELECT = 0
JOY_DEADZONE = 0.5
JOY_DELAY = 300  # ms
JOY_DOWN = (0, -1)
JOY_LEFT = (-1, 0)
JOY_REPEAT = 100  # ms
JOY_RIGHT = (1, 0)
JOY_UP = (0, 1)

# Keyboard events
KEY_APPLY = _locals.K_RETURN
KEY_BACK = _locals.K_BACKSPACE
KEY_CLOSE_MENU = _locals.K_ESCAPE
KEY_LEFT = _locals.K_LEFT
KEY_MOVE_DOWN = _locals.K_UP
KEY_MOVE_UP = _locals.K_DOWN  # Consider keys are "inverted"
KEY_RIGHT = _locals.K_RIGHT
KEY_TAB = _locals.K_TAB


# noinspection PyUnusedLocal
class Controller(object):
    """
    Controller class. Accepts any object and provides functions to handle each
    event.
    """
    joy_delay: int
    joy_repeat: int

    def __init__(self) -> None:
        self.joy_delay = JOY_DELAY
        self.joy_repeat = JOY_REPEAT

    @staticmethod
    def apply(event: EventType, widget: WidgetType) -> bool:
        """
        Accepts apply key. Requires ``pygame.KEYDOWN``.

        :param event: Event
        :param widget: Widget that accepts the event
        :return: True if event matches
        """
        return event.key == KEY_APPLY

    @staticmethod
    def back(event: EventType, widget: WidgetType) -> bool:
        """
        Accepts back key. Requires ``pygame.KEYDOWN``.

        :param event: Event
        :param widget: Widget that accepts the event
        :return: True if event matches
        """
        return event.key == KEY_BACK

    @staticmethod
    def close_menu(event: EventType, widget: WidgetType) -> bool:
        """
        Accepts close menu key. Requires ``pygame.KEYDOWN``.

        :param event: Event
        :param widget: Widget that accepts the event
        :return: True if event matches
        """
        return event.key == KEY_CLOSE_MENU

    @staticmethod
    def delete(event: EventType, widget: WidgetType) -> bool:
        """
        Accepts delete key. Requires ``pygame.KEYDOWN``.

        :param event: Event
        :param widget: Widget that accepts the event
        :return: True if event matches
        """
        return event.key == _locals.K_DELETE

    @staticmethod
    def end(event: EventType, widget: WidgetType) -> bool:
        """
        Accepts end key. Requires ``pygame.KEYDOWN``.

        :param event: Event
        :param widget: Widget that accepts the event
        :return: True if event matches
        """
        return event.key == _locals.K_END

    @staticmethod
    def escape(event: EventType, widget: WidgetType) -> bool:
        """
        Accepts escape key. Requires ``pygame.KEYDOWN``.

        :param event: Event
        :param widget: Widget that accepts the event
        :return: True if event matches
        """
        return event.key == _locals.K_ESCAPE

    @staticmethod
    def home(event: EventType, widget: WidgetType) -> bool:
        """
        Accepts home key. Requires ``pygame.KEYDOWN``.

        :param event: Event
        :param widget: Widget that accepts the event
        :return: True if event matches
        """
        return event.key == _locals.K_HOME

    @staticmethod
    def joy_axis_x_left(event: EventType, widget: WidgetType) -> bool:
        """
        Accepts joy movement on x-axis (left direction). Requires ``pygame.JOYAXISMOTION``.

        :param event: Event
        :param widget: Widget that accepts the event
        :return: True if event matches
        """
        return event.axis == JOY_AXIS_X and event.value < -JOY_DEADZONE

    @staticmethod
    def joy_axis_x_right(event: EventType, widget: WidgetType) -> bool:
        """
        Accepts joy movement on x-axis (right direction). Requires ``pygame.JOYAXISMOTION``.

        :param event: Event
        :param widget: Widget that accepts the event
        :return: True if event matches
        """
        return event.axis == JOY_AXIS_X and event.value > JOY_DEADZONE

    @staticmethod
    def joy_axis_y_down(event: EventType, widget: WidgetType) -> bool:
        """
        Accepts joy movement on y-axis (down direction). Requires ``pygame.JOYAXISMOTION``.

        :param event: Event
        :param widget: Widget that accepts the event
        :return: True if event matches
        """
        return event.axis == JOY_AXIS_Y and event.value > JOY_DEADZONE

    @staticmethod
    def joy_axis_y_up(event: EventType, widget: WidgetType) -> bool:
        """
        Accepts joy movement on y-axis (up direction). Requires ``pygame.JOYAXISMOTION``.

        :param event: Event
        :param widget: Widget that accepts the event
        :return: True if event matches
        """
        return event.axis == JOY_AXIS_Y and event.value < -JOY_DEADZONE

    @staticmethod
    def joy_back(event: EventType, widget: WidgetType) -> bool:
        """
        Accepts joy back button. Requires ``pygame.JOYBUTTONDOWN``.

        :param event: Event
        :param widget: Widget that accepts the event
        :return: True if event matches
        """
        return event.button == JOY_BUTTON_BACK

    @staticmethod
    def joy_down(event: EventType, widget: WidgetType) -> bool:
        """
        Accepts joy movement to down direction. Requires ``pygame.JOYHATMOTION``.

        :param event: Event
        :param widget: Widget that accepts the event
        :return: True if event matches
        """
        return event.value == JOY_DOWN

    @staticmethod
    def joy_left(event: EventType, widget: WidgetType) -> bool:
        """
        Accepts joy movement to left direction. Requires ``pygame.JOYHATMOTION``.

        :param event: Event
        :param widget: Widget that accepts the event
        :return: True if event matches
        """
        return event.value == JOY_LEFT

    @staticmethod
    def joy_right(event: EventType, widget: WidgetType) -> bool:
        """
        Accepts joy movement to right direction. Requires ``pygame.JOYHATMOTION``.

        :param event: Event
        :param widget: Widget that accepts the event
        :return: True if event matches
        """
        return event.value == JOY_RIGHT

    @staticmethod
    def joy_select(event: EventType, widget: WidgetType) -> bool:
        """
        Accepts joy select button. Also used for apply(). Requires ``pygame.JOYBUTTONDOWN``.

        :param event: Event
        :param widget: Widget that accepts the event
        :return: True if event matches
        """
        return event.button == JOY_BUTTON_SELECT

    @staticmethod
    def joy_up(event: EventType, widget: WidgetType) -> bool:
        """
        Accepts joy movement to up direction. Requires ``pygame.JOYHATMOTION``.

        :param event: Event
        :param widget: Widget that accepts the event
        :return: True if event matches
        """
        return event.value == JOY_UP

    @staticmethod
    def left(event: EventType, widget: WidgetType) -> bool:
        """
        Accepts left key. Requires ``pygame.KEYDOWN``.

        :param event: Event
        :param widget: Widget that accepts the event
        :return: True if event matches
        """
        return event.key == KEY_LEFT

    @staticmethod
    def move_down(event: EventType, widget: WidgetType) -> bool:
        """
        Accepts move down key. Requires ``pygame.KEYDOWN``.

        :param event: Event
        :param widget: Widget that accepts the event
        :return: True if event matches
        """
        return event.key == KEY_MOVE_DOWN

    @staticmethod
    def move_up(event: EventType, widget: WidgetType) -> bool:
        """
        Accepts move up key. Requires ``pygame.KEYDOWN``.

        :param event: Event
        :param widget: Widget that accepts the event
        :return: True if event matches
        """
        return event.key == KEY_MOVE_UP

    @staticmethod
    def right(event: EventType, widget: WidgetType) -> bool:
        """
        Accepts right key. Requires ``pygame.KEYDOWN``.

        :param event: Event
        :param widget: Widget that accepts the event
        :return: True if event matches
        """
        return event.key == KEY_RIGHT

    @staticmethod
    def tab(event: EventType, widget: WidgetType) -> bool:
        """
        Accepts tab key. Requires ``pygame.KEYDOWN``.

        :param event: Event
        :param widget: Widget that accepts the event
        :return: True if event matches
        """
        return event.key == KEY_TAB



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

NONE WIDGET
None widget definition.
"""

__all__ = [
    'NoneWidget',
    'NoneWidgetManager'
]

import pygame
import pygame_menu

from abc import ABC
from pygame_menu.utils import make_surface
from pygame_menu.widgets.core.widget import Widget, WidgetTransformationNotImplemented, \
    AbstractWidgetManager

from pygame_menu._types import Optional, NumberType, EventVectorType


# noinspection PyMissingOrEmptyDocstring
class NoneWidget(Widget):
    """
    None widget. Useful if used for filling column/row layout. None widget don't
    accept values, padding, margin, cursors, position, sound, controls, and
    cannot be selected. Also, none widget cannot accept callbacks, except draw
    and update callbacks.

    .. note::

        NoneWidget does not accept any transformation.

    :param widget_id: ID of the widget
    """

    def __init__(
        self,
        widget_id: str = ''
    ) -> None:
        super(NoneWidget, self).__init__(widget_id=widget_id)
        self.is_selectable = False
        self._surface = make_surface(0, 0, alpha=True)

    def _apply_font(self) -> None:
        pass

    def set_padding(self, *args, **kwargs) -> 'NoneWidget':
        return self

    def get_selected_time(self) -> NumberType:
        return 0

    def set_title(self, *args, **kwargs) -> 'NoneWidget':
        return self

    def get_rect(self, *args, **kwargs) -> 'pygame.Rect':
        return pygame.Rect(0, 0, 0, 0)

    def set_background_color(self, *args, **kwargs) -> 'NoneWidget':
        return self

    def _draw_background_color(self, *args, **kwargs) -> None:
        pass

    def _draw_border(self, *args, **kwargs) -> None:
        pass

    def set_selection_effect(self, *args, **kwargs) -> 'NoneWidget':
        return self

    def apply(self, *args) -> None:
        pass

    def change(self, *args) -> None:
        pass

    def _draw(self, *args, **kwargs) -> None:
        pass

    def _render(self, *args, **kwargs) -> Optional[bool]:
        pass

    def set_margin(self, *args, **kwargs) -> 'NoneWidget':
        return self

    def _apply_transforms(self, *args, **kwargs) -> None:
        pass

    def set_font(self, *args, **kwargs) -> 'NoneWidget':
        return self

    def update_font(self, *args, **kwargs) -> 'NoneWidget':
        return self

    def set_position(self, *args, **kwargs) -> 'NoneWidget':
        return self

    def scale(self, *args, **kwargs) -> 'NoneWidget':
        raise WidgetTransformationNotImplemented()

    def resize(self, *args, **kwargs) -> 'NoneWidget':
        raise WidgetTransformationNotImplemented()

    def set_max_width(self, *args, **kwargs) -> 'NoneWidget':
        raise WidgetTransformationNotImplemented()

    def set_max_height(self, *args, **kwargs) -> 'NoneWidget':
        raise WidgetTransformationNotImplemented()

    def rotate(self, *args, **kwargs) -> 'NoneWidget':
        raise WidgetTransformationNotImplemented()

    def flip(self, *args, **kwargs) -> 'NoneWidget':
        raise WidgetTransformationNotImplemented()

    def translate(self, *args, **kwargs) -> 'NoneWidget':
        raise WidgetTransformationNotImplemented()

    def select(self, *args, **kwargs) -> 'NoneWidget':
        return self

    def set_font_shadow(self, *args, **kwargs) -> 'NoneWidget':
        return self

    def set_sound(self, *args, **kwargs) -> 'NoneWidget':
        return self

    def set_cursor(self, *args, **kwargs) -> 'NoneWidget':
        return self

    def set_controls(self, *args, **kwargs) -> 'NoneWidget':
        return self

    def set_border(self, *args, **kwargs) -> 'NoneWidget':
        return self

    def _check_mouseover(self, *args, **kwargs) -> bool:
        self._mouseover = False
        return False

    def mouseleave(self, *args, **kwargs) -> 'NoneWidget':
        return self

    def mouseover(self, *args, **kwargs) -> 'NoneWidget':
        return self

    def set_onchange(self, *args, **kwargs) -> 'NoneWidget':
        self._onchange = None
        return self

    def set_onreturn(self, *args, **kwargs) -> 'NoneWidget':
        self._onreturn = None
        return self

    def set_onmouseleave(self, *args, **kwargs) -> 'NoneWidget':
        self._onmouseleave = None
        return self

    def set_onmouseover(self, *args, **kwargs) -> 'NoneWidget':
        self._onmouseover = None
        return self

    def set_onselect(self, *args, **kwargs) -> 'NoneWidget':
        self._onselect = None
        return self

    def set_tab_size(self, *args, **kwargs) -> 'NoneWidget':
        return self

    def shadow(self, *args, **kwargs) -> 'NoneWidget':
        return self

    def update(self, events: EventVectorType) -> bool:
        self.apply_update_callbacks(events)
        return False


class NoneWidgetManager(AbstractWidgetManager, ABC):
    """
    NoneWidget manager.
    """

    def none_widget(
        self,
        widget_id: str = ''
    ) -> 'pygame_menu.widgets.NoneWidget':
        """
        Add a none widget to the Menu.

        .. note::

            This widget is useful to fill column/rows layout without compromising
            any visuals. Also, it can be used to store information or even to add
            a ``draw_callback`` function to it for being called on each Menu draw.

        .. note::

            This is applied only to the base Menu (not the currently displayed,
            stored in ``_current`` pointer); for such behaviour apply to
            :py:meth:`pygame_menu.menu.Menu.get_current` object.

        :param widget_id: Widget ID
        :return: Widget object
        :rtype: :py:class:`pygame_menu.widgets.NoneWidget`
        """
        attributes = self._filter_widget_attributes({})

        widget = NoneWidget(widget_id=widget_id)
        self._configure_widget(widget=widget, **attributes)
        self._append_widget(widget)

        return widget
